# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl import app, logging

import neurst.utils.flags_core as flags_core
from neurst.data.datasets import Dataset, build_dataset
from neurst.exps import BaseExperiment, build_exp
from neurst.models import BaseModel
from neurst.tasks import Task, build_task
from neurst.training import training_utils
from neurst.utils.configurable import ModelConfigs, deep_merge_dict, load_from_config_path, yaml_load_checking
from neurst.utils.hparams_sets import get_hyper_parameters
from neurst.utils.misc import flatten_string_list

FLAG_LIST = [
    flags_core.Flag("distribution_strategy", dtype=flags_core.Flag.TYPE.STRING, default="mirrored",
                    help="The distribution strategy."),
    flags_core.Flag("dtype", dtype=flags_core.Flag.TYPE.STRING, default="float16",
                    help="The computation type of the whole model."),
    flags_core.Flag("enable_check_numerics", dtype=flags_core.Flag.TYPE.BOOLEAN, default=None,
                    help="Whether to open the tf.debugging.enable_check_numerics. "
                         "Note that this may lower down the training speed."),
    flags_core.Flag("enable_xla", dtype=flags_core.Flag.TYPE.BOOLEAN, default=None,
                    help="Whether to enable XLA for training."),
    flags_core.Flag("hparams_set", dtype=flags_core.Flag.TYPE.STRING,
                    help="A string indicating a set of pre-defined hyper-parameters, "
                         "e.g. transformer_base, transformer_big or transformer_768_16e_3d."),
    flags_core.Flag("model_dir", dtype=flags_core.Flag.TYPE.STRING,
                    help="The path to the checkpoint for saving and loading."),
    flags_core.ModuleFlag(BaseExperiment.REGISTRY_NAME, help="The program."),
    flags_core.ModuleFlag(Task.REGISTRY_NAME, help="The binding task."),
    flags_core.ModuleFlag(BaseModel.REGISTRY_NAME, help="The model."),
    flags_core.ModuleFlag(Dataset.REGISTRY_NAME, help="The dataset."),
]


def _pre_load_args(args):
    cfg_file_args = yaml_load_checking(load_from_config_path(
        flatten_string_list(getattr(args, flags_core.DEFAULT_CONFIG_FLAG.name))))
    model_dirs = flatten_string_list(args.model_dir or cfg_file_args.get("model_dir", None))
    hparams_set = args.hparams_set
    if hparams_set is None:
        hparams_set = cfg_file_args.get("hparams_set", None)
    predefined_parameters = get_hyper_parameters(hparams_set)
    formatted_parameters = {}
    if "model.class" in predefined_parameters:
        formatted_parameters["model.class"] = predefined_parameters.pop("model.class")
    if "model" in predefined_parameters:
        formatted_parameters["model"] = predefined_parameters.pop("model")
    if "model.params" in predefined_parameters:
        formatted_parameters["model.params"] = predefined_parameters.pop("model.params")
    if len(predefined_parameters) > 0:
        formatted_parameters["entry.params"] = predefined_parameters

    try:
        model_cfgs = ModelConfigs.load(model_dirs[0])
        return deep_merge_dict(deep_merge_dict(
            model_cfgs, formatted_parameters), cfg_file_args)
    except Exception:
        return deep_merge_dict(formatted_parameters, cfg_file_args)


def run_experiment(args, remaining_argv):
    strategy = training_utils.handle_distribution_strategy(args["distribution_strategy"])
    flags_core.verbose_flags(FLAG_LIST, args, remaining_argv)
    training_utils.startup_env(
        dtype=args["dtype"],
        enable_check_numerics=args["enable_check_numerics"],
        enable_xla=args["enable_xla"])

    # create exps: trainer, evaluator or ...
    with training_utils.get_strategy_scope(strategy):
        task = build_task(args)
        custom_dataset = build_dataset(args)
        try:
            model = task.build_model(args)
        except AttributeError:
            model = None
        entry = build_exp(args,
                          strategy=strategy,
                          model=model,
                          task=task,
                          model_dir=args["model_dir"],
                          custom_dataset=custom_dataset)
    entry.run()


def _main(_):
    # define and parse program flags
    arg_parser = flags_core.define_flags(FLAG_LIST)
    args, remaining_argv = flags_core.intelligent_parse_flags(FLAG_LIST, arg_parser, _pre_load_args)
    args, remaining_argv = flags_core.extend_define_and_parse(
        BaseExperiment.REGISTRY_NAME, args, remaining_argv)
    if args["entry.class"] is None:
        raise ValueError("Must provide entry/entry.class.")
    run_experiment(args, remaining_argv)


def cli_main():
    logging.set_verbosity(logging.INFO)
    app.run(_main, argv=["pseudo.py"])


if __name__ == "__main__":
    cli_main()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from distutils.version import LooseVersion

import tensorflow as tf
from absl import logging
from tensorflow.keras.mixed_precision.experimental import LossScaleOptimizer
from tensorflow.python.keras import backend
from tensorflow.python.training.experimental import loss_scale as loss_scale_module

from neurst.criterions import Criterion, build_criterion
from neurst.data.dataset_utils import map_data_for_keras
from neurst.data.datasets.multiple_dataset import MultipleDataset
from neurst.exps import BaseExperiment, register_exp
from neurst.models.model_utils import summary_model_variables
from neurst.optimizers import OPTIMIZER_REGISTRY_NAME, build_optimizer
from neurst.optimizers.schedules import LR_SCHEDULE_REGISTRY_NAME, build_lr_schedule
from neurst.training import (CustomCheckpointCallback, LearningRateScheduler, MetricReductionCallback, Validator,
                             build_validator, training_utils)
from neurst.training.accumgrad_keras_model import AccumgradKerasModel
from neurst.utils import compat
from neurst.utils.checkpoints import restore_checkpoint_if_possible, restore_checkpoint_if_possible_v2
from neurst.utils.flags_core import Flag, ModuleFlag
from neurst.utils.misc import flatten_string_list


def _handle_fp16_and_distributed_optimizer(optimizer, lr_schedule, hvd_backend=None):
    if hvd_backend == "horovod":
        import horovod.tensorflow.keras as hvd
        from horovod.tensorflow import Compression
    elif hvd_backend == "byteps":
        import byteps.tensorflow.keras as hvd
        from byteps.tensorflow import Compression

    if hvd_backend:
        compression = Compression.none
        if compat.CUSTOM_GLOBAL_FLOATX == "float16":
            compression = Compression.fp16

    if lr_schedule is not None and hvd_backend is None:
        # TODO(ZhaoChengqi): pay attention to API changes
        optimizer._set_hyper("learning_rate", lr_schedule)
    # specify the following scenario
    # there is a bug under TF2.3+Horovod+fp16+XLA
    if compat.CUSTOM_GLOBAL_FLOATX == "float16":
        logging.info("NOTICE: using revised DynamicLossScale under fp16")
        revised_loss_scale = training_utils.RevisedDynamicLossScale()
        if hvd_backend:
            opt = LossScaleOptimizer(optimizer, loss_scale=1)
            opt = hvd.DistributedOptimizer(opt, compression=compression, sparse_as_dense=True)
            opt._loss_scale = revised_loss_scale
            for weight in loss_scale_module.get_loss_scale_weights(opt._loss_scale):
                backend.track_variable(weight)
            opt._track_trackable(opt._loss_scale, 'loss_scale', overwrite=True)
        else:
            opt = LossScaleOptimizer(optimizer, loss_scale=revised_loss_scale)
        return opt
    return optimizer


@register_exp("train")
class Trainer(BaseExperiment):
    """ Trainer for all tasks. """

    def __init__(self, args, **kwargs):
        """ Initializes a util class for training neural models. """
        super(Trainer, self).__init__(**kwargs)
        self._tb_log_dir = args["tb_log_dir"]
        self._train_steps = args["train_steps"]
        self._summary_steps = args["summary_steps"]
        self._save_checkpoint_steps = args["save_checkpoint_steps"]
        self._checkpoints_max_to_keep = args["checkpoints_max_to_keep"]
        self._initial_global_step = args["initial_global_step"]
        self._pretrain_variable_pattern = args["pretrain_variable_pattern"]
        if args["pretrain_model"] and isinstance(args["pretrain_model"][0], dict):
            self._pretrain_v2 = True
            self._pretrain_model = args["pretrain_model"]
            if self._pretrain_variable_pattern:
                logging.info("Using pretrain model v2 and ignoring pretrain_variable_pattern: "
                             f"{self._pretrain_variable_pattern}")
        else:
            self._pretrain_v2 = False
            self._pretrain_model = flatten_string_list(args["pretrain_model"])
            if self._pretrain_model and self._pretrain_variable_pattern is None:
                self._pretrain_variable_pattern = [None] * len(self._pretrain_model)
            assert ((self._pretrain_model is None and self._pretrain_variable_pattern is None)
                    or len(self._pretrain_model) == len(self._pretrain_variable_pattern)
                    or len(self._pretrain_model) == 1), (
                "`pretrain_variable_pattern` must match with `pretrain_model`.")
            if self._pretrain_model is not None and self._pretrain_variable_pattern is None:
                self._pretrain_variable_pattern = [None] * len(self._pretrain_model)
        self._update_cycle = args["update_cycle"]
        self._clip_value = args["clip_value"]
        self._clip_norm = args["clip_norm"]
        self._hvd_backend = self.strategy if self.strategy in ["byteps", "horovod"] else None
        with training_utils.get_strategy_scope(self.strategy):
            self._criterion = build_criterion(args)
            self._criterion.set_model(self.model)
            self._lr_schedule_args = args
            self._optimizer = build_optimizer(args)
            assert self._optimizer is not None, "optimizer parameters must be provided for training."
            # self._optimizer = _handle_fp16_and_distributed_optimizer(optimizer, self._lr_schedule, self._hvd_backend)
        self._validator = build_validator(args)
        self._experimental_count_batch_num = args["experimental_count_batch_num"]
        self._freeze_variables = args["freeze_variables"]

    @staticmethod
    def class_or_method_args():
        return [
            ModuleFlag(Criterion.REGISTRY_NAME, help="The criterion for training or evaluation."),
            ModuleFlag(OPTIMIZER_REGISTRY_NAME, help="The optimizer for training."),
            ModuleFlag(LR_SCHEDULE_REGISTRY_NAME, help="The learning schedule for training."),
            ModuleFlag(Validator.REGISTRY_NAME, help="The validation process while training."),
            Flag("tb_log_dir", dtype=Flag.TYPE.STRING, default=None,
                 help="The path to store tensorboard summary, or `model_dir`/train by default."),
            Flag("train_steps", dtype=Flag.TYPE.INTEGER, default=10000000,
                 help="The maximum steps for training loop."),
            Flag("summary_steps", dtype=Flag.TYPE.INTEGER, default=200,
                 help="Doing summary(logging & tensorboard) this every steps."),
            Flag("save_checkpoint_steps", dtype=Flag.TYPE.INTEGER, default=1000,
                 help="Saving checkpoints this every steps."),
            Flag("checkpoints_max_to_keep", dtype=Flag.TYPE.INTEGER, default=8,
                 help="The maximum checkpoints to be kept."),
            Flag("initial_global_step", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The manually specified initial global step."),
            Flag("pretrain_model", dtype=Flag.TYPE.STRING, default=None, multiple=True,
                 help="The path to a pretrained model directory(a seq2seq model, bert model, etc.). "
                      "(V2) Or a json/yaml-like dict string indicating pretrained models from "
                      "either neurst checkpoints or publicly available models converted "
                      "by neurst.utils.converters. Each entry has the elements: "
                      "path, model_name, from_prefix, to_prefix, name_pattern. "
                      "Multiple pretrain models are also available."),
            Flag("pretrain_variable_pattern", dtype=Flag.TYPE.STRING, default=None, multiple=True,
                 help="One can restore specified variables in the `pretrain_model` by this regular expression."
                      "Multiple pattern are also available, but must match to `pretrain_model`."),
            Flag("update_cycle", dtype=Flag.TYPE.INTEGER, default=1,
                 help="Training step with this many batches (Gradient Accumulation)."),
            Flag("clip_value", dtype=Flag.TYPE.FLOAT, default=None, help="Gradient clipping by value."),
            Flag("clip_norm", dtype=Flag.TYPE.FLOAT, default=None, help="Gradient clipping by norm."),
            Flag("experimental_count_batch_num", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Pre-scan the dataset for training and count the number of batches."),
            Flag("freeze_variables", dtype=Flag.TYPE.STRING, default=None,
                 help="Variables whose names are matched with this regex will be freezed.")
        ]

    def _restore_ckpt_or_pretrain(self):
        """ restoring checkpoint from model_dir or pretrain_model dir. """
        stat = restore_checkpoint_if_possible(self.model, self.model_dir)
        continue_training = False
        if stat:
            logging.info(f"Successfully restoring checkpoint from model_dir={self.model_dir}")
            continue_training = True
        else:
            logging.info(f"No checkpoint restored from model_dir={self.model_dir}")
            if self._pretrain_model:
                if self._pretrain_v2:
                    for pt in self._pretrain_model:
                        logging.info(f"Trying to restore from pretrain_model={pt}")
                        logging.info("NOTE THAT, one must first check the variable names in this checkpoint, "
                                     "otherwise no variables will be restored.")
                        restore_checkpoint_if_possible_v2(self.model, **pt)
                else:
                    for pt, pt_varname in zip(self._pretrain_model, self._pretrain_variable_pattern):
                        logging.info(f"Trying to restore from pretrain_model={pt}")
                        logging.info("NOTE THAT, one must first check the variable names in this checkpoint, "
                                     "otherwise no variables will be restored.")
                        restore_checkpoint_if_possible(self.model, pt, var_name_pattern=pt_varname)

        if self._initial_global_step is None and continue_training:
            _step = compat.hack_global_step(self.model_dir)
            if _step:
                compat.register_initial_step(_step or 0)  # must do this before creating optimizer and training
                logging.info(f"Restored initial global step={_step}")
        else:
            compat.register_initial_step(self._initial_global_step or 0)

    def run(self):
        """ Training a neural model.

        Step 1: Create training model
        Step 2: Restore checkpoint/pretrain model/global_step if exists.
        Step 3: Fetch training data.
        Step 5: Fetch training training.
        Step 6: TRAIN!!!
        """
        if self._hvd_backend == "horovod":
            import horovod.tensorflow.keras as hvd
        elif self._hvd_backend == "byteps":
            import byteps.tensorflow.keras as hvd

        tfds = training_utils.build_datasets(compat.ModeKeys.TRAIN, self.strategy,
                                             self.custom_dataset, self.task)
        if isinstance(self.custom_dataset, MultipleDataset):
            _tfds = None
            for _, ds in tfds.items():
                if _tfds is None:
                    _tfds = ds
                else:
                    _tfds = _tfds.concatenate(ds)
            tfds = _tfds
        tfds = tfds.prefetch(tf.data.experimental.AUTOTUNE)
        # Step 1: create a model
        with training_utils.get_strategy_scope(self.strategy):
            inps = self.task.create_inputs(compat.ModeKeys.TRAIN)
            formatted_inps = self.task.example_to_input(inps, compat.ModeKeys.TRAIN)
            model_out = self.model(formatted_inps, is_training=True)
            for metric_layer in self.task.build_metric_layer():
                model_out = metric_layer([formatted_inps, model_out])
            if LooseVersion(tf.__version__) >= LooseVersion("2.3"):
                keras_model = AccumgradKerasModel(inps, model_out,
                                                  update_cycle=self._update_cycle,
                                                  clip_value=self._clip_value,
                                                  clip_norm=self._clip_norm,
                                                  freeze_variables=self._freeze_variables)
            else:
                logging.info(f"Warning: ignore update_cycle={self._update_cycle}, grad_clip={self._grad_clip} "
                             f"when TensorFlow version < 2.3.")
                keras_model = tf.keras.Model(inps, model_out)
            loss = self._criterion.reduce_loss(formatted_inps, model_out)
            if isinstance(loss, (list, tuple, tf.Tensor)):
                keras_model.add_loss(loss)
            elif isinstance(loss, dict):
                for _name, _loss in loss.items():
                    keras_model.add_loss(_loss)
                    keras_model.add_metric(_loss, name=_name + "_mean", aggregation="mean")
            else:
                raise ValueError("criterion.reduce_loss returns "
                                 "unsupported value of type: {}".format(type(loss)))
            self._restore_ckpt_or_pretrain()
            self._lr_schedule = build_lr_schedule(self._lr_schedule_args)
            self._optimizer = _handle_fp16_and_distributed_optimizer(
                self._optimizer, self._lr_schedule, self._hvd_backend)
            if self._hvd_backend is None:
                keras_model.compile(self._optimizer)
            else:
                # NOTE: we already add Horovod DistributedOptimizer in `_handle_fp16_and_distributed_optimizer`.
                # Horovod: Specify `experimental_run_tf_function=False` to ensure TensorFlow
                # uses hvd.DistributedOptimizer() to compute gradients.
                keras_model.compile(self._optimizer, experimental_run_tf_function=False)
            keras_model.summary()
            summary_model_variables(self.model, self._freeze_variables)
        # initialize the checkpoint manager
        _ = compat.get_saver_or_default(self.model, self.model_dir, max_to_keep=self._checkpoints_max_to_keep)
        # build training training
        if not self._tb_log_dir:
            self._tb_log_dir = os.path.join(self.model_dir, "train")

        training_callbacks = [MetricReductionCallback(self.strategy, self._summary_steps, self._tb_log_dir,
                                                      device="GPU:0", lr_schedule=self._lr_schedule)]
        if self._hvd_backend is None or hvd.rank() == 0:
            training_callbacks.append(
                CustomCheckpointCallback(self.task.model_configs(self.model),
                                         save_checkpoint_steps=self._save_checkpoint_steps))
            if self._validator is not None:
                training_callbacks.append(self._validator.build(self.strategy, self.task, self.model))
        if self._hvd_backend is not None:
            # Horovod: average metrics among workers at the end of every epoch.
            #
            # Note: This callback must be in the list before the ReduceLROnPlateau,
            # TensorBoard or other metrics-based training.
            # NOTE!!! HERE we already integrate the metric averaging behaviour into the MetricReductionCallback.
            # training_callbacks.insert(0, hvd.callbacks.MetricAverageCallback(device="GPU:0"))

            # Horovod: broadcast initial variable states from rank 0 to all other processes.
            # This is necessary to ensure consistent initialization of all workers when
            # training is started with random weights or restored from a checkpoint.
            training_callbacks.insert(0, hvd.callbacks.BroadcastGlobalVariablesCallback(0, device="GPU:0"))
            if self._lr_schedule is not None:
                training_callbacks.append(LearningRateScheduler(self._lr_schedule))

        if self._experimental_count_batch_num:
            logging.info("Scanning the dataset......")
            iterator = iter(training_utils.maybe_distribution_dataset(self.strategy, tfds))
            cnt = 0
            for _ in iterator:
                cnt += 1
            logging.info(f"Total {cnt} batches per EPOCH.")

        history = keras_model.fit(
            map_data_for_keras(tfds.repeat()),
            initial_epoch=0,
            epochs=1,
            steps_per_epoch=self._train_steps,  # * args["update_cycle"],
            verbose=2,
            callbacks=training_callbacks)
        logging.info(history.history)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import json
import random

import tensorflow as tf


class PaddingMode(object):
    DEFAULT = 1
    EOS_AS_PADDING = 2


class Vocab(object):
    def __init__(self, tokens, extra_tokens=None, lowercase=False):
        """ Initialize vocabulary

        Args:
            tokens: A list of word tokens.
            extra_tokens: A list of word tokens appended to the end of the vocab, which
                will not be affected by the `lowercase` option.
            lowercase: whether transfer to lowercase
        """
        assert isinstance(tokens, list), (
            "`tokens` must be a list of string tokens")
        if lowercase:
            uniq_tokens = []
            for token in tokens:
                token = token.lower()
                if token not in uniq_tokens:
                    uniq_tokens.append(token)
            tokens = uniq_tokens
        self._token_list = copy.deepcopy(tokens)
        if isinstance(extra_tokens, list):
            for token in extra_tokens:
                if token not in self._token_list:
                    self._token_list.append(token)
        self._token_to_id_dict = dict([
            (w, i) for i, w in enumerate(self._token_list)])
        self._lowercase = lowercase
        self._extra_tokens = extra_tokens

    @property
    def tokens(self):
        return self._token_list

    @property
    def vocab_size(self):
        return len(self._token_list)

    def add_word(self, w, lowercase=False):
        """ Adds word to the end of the vocabulary. """
        if lowercase and self._lowercase:
            w = w.lower()
        if w not in self._token_list:
            self._token_list.append(w)
            self._token_to_id_dict[w] = len(self._token_list) - 1

    @staticmethod
    def load_tokens(vocab_path=None, tokens=None):
        if not ((vocab_path is None) ^ (tokens is None)):
            raise ValueError("Either `vocab_path` or `tokens` should be provided.")
        if vocab_path:
            with tf.io.gfile.GFile(vocab_path) as f:
                if vocab_path.endswith(".json"):
                    tokens = list(json.load(f).keys())
                else:
                    tokens = [line.strip("\n") for line in f]
        cleaned_tokens = []
        assert isinstance(tokens, list)
        for word in tokens:
            if (len(word) > 1 and ((word.startswith("'") and word.endswith("'"))
                                   or (word.startswith('"') and word.endswith('"')))):
                word = word[1:-1]
            else:
                if word.strip() != "":
                    word = word.strip().split()[0]
            if word == "":
                continue
            cleaned_tokens.append(word)
        return cleaned_tokens

    @staticmethod
    def get_unique(codebook, token):
        while token in codebook:
            token += str(random.choice(list(range(0, 10))))
        return token

    @staticmethod
    def load_from_file(vocab_path, extra_tokens=None, lowercase=False):
        """ Init vocabulary from file

        Args:
            vocab_path: A file path of the vocabulary
            extra_tokens: extra tokens appended to the end
            lowercase: whether transfer to lowercase

        Returns: A `Vocab` instance
        """
        return Vocab(Vocab.load_tokens(vocab_path, tokens=None), extra_tokens, lowercase)

    def map_token_to_id(self, tokens, unknown_default=None):
        """ Map a token to int id, if unk return unk_id

        Args:
            tokens: a word token or a list of tokens
            unknown_default: the default id if token not exists

        Returns: int id or a list of int ids
        """

        def _map(t):
            if self._lowercase and t not in self._extra_tokens:
                t = t.lower()
            try:
                return self._token_to_id_dict[t]
            except KeyError:
                return unknown_default

        if isinstance(tokens, list):
            return [_map(t) for t in tokens]
        assert isinstance(tokens, str)
        return _map(tokens)

    def map_id_to_token(self, ids):
        """ Map an integer(s) to word token(s)

        Args:
            ids: An integer or a list of integers

        Returns: a string token or list of strings
        """
        if isinstance(ids, list):
            return [self._token_list[i] for i in ids]
        assert isinstance(ids, int)
        return self._token_list[ids]



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import time
import traceback

import tensorflow as tf
from absl import logging

from neurst.exps import BaseExperiment, register_exp
from neurst.tasks import build_task
from neurst.training import Validator, build_validator
from neurst.utils import compat
from neurst.utils.configurable import ModelConfigs
from neurst.utils.flags_core import Flag, ModuleFlag


@register_exp
class Validation(BaseExperiment):
    """ Validation for all tasks during training. """

    def __init__(self, args, **kwargs):
        """ Initializes a util class for vaidation. """
        super(Validation, self).__init__(**kwargs)
        self._tb_log_dir = args["tb_log_dir"]
        self._waiting_interval = args["waiting_interval"]
        self._maximum_waiting_time = args["maximum_waiting_time"]
        self._validator = build_validator(args)

    @staticmethod
    def class_or_method_args():
        return [
            Flag("tb_log_dir", dtype=Flag.TYPE.STRING, default=None,
                 help="The path to store tensorboard summary, or `model_dir`/validation by default."),
            Flag("waiting_interval", dtype=Flag.TYPE.INTEGER, default=120,
                 help="The waiting interval between two evaluation steps."),
            Flag("maximum_waiting_time", dtype=Flag.TYPE.INTEGER, default=3600,
                 help="The maximum waiting time(in seconds)."),
            ModuleFlag(Validator.REGISTRY_NAME, help="The validation process during training."),
        ]

    def run(self):
        """ Repeats to call validator's validate function if new checkponts are observed.

        Step 1: Build model.
        Step 2: Fetch training status.
        while True:
            Step 3: Restore checkpoints.
            Step 4: Validate.
        """
        if self.task is None or self.model is None:
            model_cfg_waiting_rounds = self._maximum_waiting_time // self._waiting_interval
            for i in range(model_cfg_waiting_rounds):
                try:
                    args = ModelConfigs.load(self._model_dir)
                    break
                except FileNotFoundError:
                    logging.info(f"Fail to load model configs from directory: {self.model_dir}. "
                                 f"Wait for another {self._waiting_interval}s, "
                                 f"patience={model_cfg_waiting_rounds - 1 - i}.")
                    time.sleep(self._waiting_interval)
            self._task = build_task(args)
            self._model = self.task.build_model(args)
        # initialize the checkpoint manager
        saver = compat.get_saver_or_default(self.model, self.model_dir)
        # enable tensorboard
        if self._tb_log_dir is None:
            self._tb_log_dir = os.path.join(self.model_dir, "validation_{}".format(int(time.time())))
        file_writer = tf.summary.create_file_writer(self._tb_log_dir)
        file_writer.set_as_default()
        # create training
        self._validator.build(self.strategy, self.task, self.model)
        last_triggered_step = None
        accumulated_waiting_time = 0
        this_waiting_interval = next_waiting_interval = self._waiting_interval
        while True:
            bad_cnt = 0
            while bad_cnt < 5:
                try:
                    ckpt_state = tf.train.get_checkpoint_state(self.model_dir)
                    break
                except ValueError:
                    bad_cnt += 1
                    time.sleep(5)
                    logging.info(traceback.format_exc())
                    if bad_cnt >= 5:
                        ckpt_state = tf.train.get_checkpoint_state(self.model_dir)

            ckpts_to_be_restore = None
            if ckpt_state is None:
                logging.info(f"No checkpoint in directory: {self.model_dir}. Please wait.")
            else:
                all_ckpts = [(t, x) for t, x in zip(ckpt_state.all_model_checkpoint_timestamps,
                                                    ckpt_state.all_model_checkpoint_paths)]
                global_steps_to_be_restore = []
                ckpts_to_be_restore = []
                for ckpt in all_ckpts[::-1]:
                    step = compat.hack_global_step(ckpt[1])
                    if last_triggered_step is None or step > last_triggered_step:
                        ckpts_to_be_restore.insert(0, ckpt)
                        global_steps_to_be_restore.insert(0, step)
                if len(ckpts_to_be_restore) > 0:
                    accumulated_waiting_time = 0
                _start_time = time.time()
                for step, (timestamp, ckpt) in zip(global_steps_to_be_restore, ckpts_to_be_restore):
                    stat = saver.restore(ckpt)
                    if not stat:
                        logging.info(f"Fail to restore checkpoint from {ckpt}. Skip...")
                        continue
                    logging.info(f"Checkpoint with global_step={step} triggered on {timestamp}")
                    self._validator.validate(step)
                    last_triggered_step = step
                this_waiting_interval = max(this_waiting_interval - int(time.time() - _start_time), 10)
                tf.summary.flush(file_writer)
            if ckpts_to_be_restore is None:
                pass
            elif len(ckpts_to_be_restore) > 1:
                this_waiting_interval = int(this_waiting_interval * 1.
                                            * (len(ckpts_to_be_restore) // 2) / len(ckpts_to_be_restore))
                next_waiting_interval = this_waiting_interval
            elif len(ckpts_to_be_restore) == 0:
                next_waiting_interval = min(int(this_waiting_interval * 4. / 3.), self._waiting_interval)
                this_waiting_interval = this_waiting_interval // 2
            accumulated_waiting_time += this_waiting_interval
            if accumulated_waiting_time > self._maximum_waiting_time:
                logging.info(f"Waited for maximum patience: {self._maximum_waiting_time}s")
                break
            time.sleep(this_waiting_interval)
            this_waiting_interval = next_waiting_interval



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from typing import Tuple

import tensorflow as tf
from absl import logging

from neurst.data import dataset_utils
from neurst.data.data_pipelines import DataPipeline, build_data_pipeline
from neurst.data.data_pipelines.transcript_data_pipeline import TranscriptDataPipeline
from neurst.data.datasets import Dataset
from neurst.layers.metric_layers.token_metric_layers import AudioFramesMetricLayer, SequenceTokenMetricLayer
from neurst.metrics import build_metric
from neurst.models import build_model
from neurst.models.model_utils import deduce_text_length
from neurst.tasks import register_task
from neurst.tasks.task import Task
from neurst.training.training_utils import minimal_multiple
from neurst.utils import compat
from neurst.utils.audio_lib import SpecAugment
from neurst.utils.configurable import deep_merge_dict
from neurst.utils.flags_core import Flag, ModuleFlag


def create_audio_bucket_boundaries(maxlen, minlen=128):
    if minlen is None:
        minlen = 128
    bounds = [minlen]
    base = minlen
    base_incr = int(2 ** ((math.log2(minlen) + 1) // 2))
    base_incr_mult = 1
    times = len(str(int(minlen)))
    while True:
        for _ in range(times):
            bounds.append(bounds[-1] + base)
            if bounds[-1] > maxlen:
                break
        base += base_incr * base_incr_mult
        base_incr_mult += 1
        if bounds[-1] > maxlen:
            break
    bounds[-1] = maxlen + 1
    return bounds


@register_task(["speech2text", "audio2text", "AudioToText"])
class SpeechToText(Task):
    """ Defines the audio to text task. """

    def __init__(self, args):
        """ Initializes the task.

        Args:
            args: A dict of model configurations.
        """
        super(SpeechToText, self).__init__(args)
        trg_data_pipeline_cls = args.get("transcript_data_pipeline.class", TranscriptDataPipeline)
        trg_data_pipeline_params = args.get("transcript_data_pipeline.params", None) or {}
        self._trg_data_pipeline = build_data_pipeline(
            trg_data_pipeline_cls, **trg_data_pipeline_params)
        self._audio_feature_dim = args["audio_feature_dim"]
        self._audio_feature_channels = args["audio_feature_channels"]
        self._specaug = SpecAugment.build(args.get("specaug", None))

    def get_config(self):
        return {
            "transcript_data_pipeline.class": self._trg_data_pipeline.__class__.__name__,
            "transcript_data_pipeline.params": self._trg_data_pipeline.get_config(),
            "audio_feature_dim": self._audio_feature_dim,
            "audio_feature_channels": self._audio_feature_channels
        }

    @staticmethod
    def class_or_method_args():
        this_args = super(SpeechToText, SpeechToText).class_or_method_args()
        this_args.extend([
            ModuleFlag("transcript_data_pipeline", DataPipeline.REGISTRY_NAME,
                       default=TranscriptDataPipeline.__name__,
                       help="The target side transcript data pipeline."),
            Flag("audio_feature_dim", dtype=Flag.TYPE.INTEGER, default=80,
                 help="The dimension of audio features."),
            Flag("audio_feature_channels", dtype=Flag.TYPE.INTEGER, default=1,
                 help="The number of channels of audio features."),
            Flag("max_src_len", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The maximum source length of training data (audio frames)."),
            Flag("min_src_bucket_boundary", dtype=Flag.TYPE.INTEGER, default=128,
                 help="The minimum source length of the training bucket (audio frames)."),
            Flag("max_trg_len", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The maximum target length of training data."),
            Flag("truncate_src", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to truncate source to max_src_len."),
            Flag("truncate_trg", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to truncate target to max_trg_len."),
            Flag("experimental_frame_transcript_ratio", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The ratio of the number of frames and its transcript for training batch bucket."),
            Flag("specaug", dtype=Flag.TYPE.STRING, default=None,
                 help="The arguments for spec augment, can be either predefined settings "
                      "like LB, LD, SM, SS... or a dict containing detailed arguments."),
            Flag("disable_batch_efficiency", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to disable the batch efficiency.")
        ])
        return this_args

    def inputs_signature(self, mode) -> Tuple[dict, dict]:
        """ Returns the input dtypes and signatures (from dataset). """
        dtypes = {"audio": tf.float32, "audio_length": tf.int64}
        # [batch, frames, feature_dim]
        signatures = {"audio": tf.TensorShape([None, None]),
                      "audio_length": tf.TensorShape([None, ])}
        if mode == compat.ModeKeys.INFER:
            return dtypes, signatures
        dtypes["transcript"] = tf.int64
        signatures["transcript"] = tf.TensorShape([None, None])
        return dtypes, signatures

    def build_model(self, args, name=None):
        """ Creates the model. """
        model = build_model(args, {"audio_feature_dim": self._audio_feature_dim,
                                   "audio_feature_channels": self._audio_feature_channels},
                            self._trg_data_pipeline.meta, name=name)
        return model

    def example_to_input(self, batch_of_data: dict, mode) -> dict:
        """ Transform the data examples to model acceptable inputs.

        Args:
            batch_of_data: A data tensor with shape [batch, ...]
            mode: The running mode.

        Returns: The input data for model.
        """
        batch = tf.shape(batch_of_data["audio"])[0]
        input_dict = {"src": tf.reshape(batch_of_data["audio"],
                                        [batch, -1, self._audio_feature_dim, self._audio_feature_channels]),
                      "src_length": batch_of_data["audio_length"]}

        target_bos = tf.tile(
            [tf.convert_to_tensor(self._trg_data_pipeline.meta["bos_id"], dtype=tf.int64)],
            [tf.shape(input_dict["src"])[0]])
        if mode == compat.ModeKeys.INFER:
            input_dict["trg_input"] = target_bos
        else:
            input_dict["trg"] = batch_of_data["transcript"]
            input_dict["trg_length"] = deduce_text_length(batch_of_data["transcript"],
                                                          self._trg_data_pipeline.meta["pad_id"],
                                                          self._trg_data_pipeline.meta["padding_mode"])
            input_dict["trg_input"] = tf.concat([tf.expand_dims(target_bos, axis=1),
                                                 batch_of_data["transcript"][:, :-1]], axis=1)
        return input_dict

    def get_data_postprocess_fn(self, mode):
        if mode == compat.ModeKeys.INFER:
            return self._trg_data_pipeline.recover
        raise ValueError("No postprocess for TRAIN/EVAL.")

    def get_data_preprocess_fn(self, mode, data_status, args=None) -> callable:
        """ Preprocess data sample according to this task.

        Args:
            args: A dict containing dataset arguments.
            mode: A ModeKeys indicating the running mode.
            data_status: The status of the data sample.

        Returns: A callable function to collate (process) a data sample.
        """
        if args is None:
            args = self._args
        else:
            args = deep_merge_dict(self._args, args, local_overwrite=False)
        trunc_audio = args.get("truncate_src", None)
        max_audio_len = args.get("max_src_len", None)
        trunc_trg = args.get("truncate_trg", None)
        max_trg_len = args.get("max_trg_len", None)

        if data_status["audio"] != compat.DataStatus.PROJECTED:
            raise RuntimeError("We recommend one to preprocess the audio in advance.")

        def _process_audio(audio):
            if trunc_audio and max_audio_len:
                audio = audio[:max_audio_len * self._audio_feature_dim * self._audio_feature_channels]
            if self._specaug is not None:
                audio = tf.reshape(
                    audio, [-1, self._audio_feature_dim * self._audio_feature_channels])
                audio = tf.reshape(self._specaug(audio), [-1])
            return audio

        def _process_and_truncate_text(text):
            if data_status["transcript"] == compat.DataStatus.RAW:
                text = self._trg_data_pipeline.process(text, is_processed=False)
            else:
                assert data_status["transcript"] == compat.DataStatus.PROJECTED
            if mode == compat.ModeKeys.TRAIN and trunc_trg and max_trg_len:
                if isinstance(text, tf.Tensor):
                    text = tf.cond(
                        tf.less_equal(tf.size(text), max_trg_len), lambda: text,
                        lambda: tf.concat([text[:(max_trg_len - 1)], text[-1:]], axis=0))
                else:
                    if len(text) > max_trg_len:
                        text = text[:(max_trg_len - 1)] + text[-1:]
            return text

        def data_proc(data, with_label):
            feature = _process_audio(data["audio"])
            ret = {"audio": feature,
                   "audio_length": tf.cast(
                       (tf.shape(feature)[0] if isinstance(feature, tf.Tensor)
                        else feature.shape[0]) // self._audio_feature_dim // self._audio_feature_channels,
                       dtype=tf.int64)}
            if with_label:
                ret["transcript"] = tf.convert_to_tensor(
                    _process_and_truncate_text(data["transcript"]), tf.int64)
            return ret

        if mode == compat.ModeKeys.INFER:
            return lambda data: data_proc(data, False)
        return lambda data: data_proc(data, True)

    def create_and_batch_tfds(self, ds: Dataset, mode,
                              args=None, num_replicas_in_sync=1) -> tf.data.Dataset:
        """ Creates a dataset according to the `mode`.

        Args:
            args: A dict containing dataset arguments.
            ds: A neurst.data.datasets.Dataset object.
            mode: A ModeKeys indicating the running mode.
            num_replicas_in_sync: The number of GPUs or other workers. We will generate global
                batches, and each global batch is equally divisible by number of replicas.

        Returns:
            A tf.data.Dataset.
        """
        if args is None:
            args = self._args
        else:
            args = deep_merge_dict(self._args, args, local_overwrite=False)
        float_zero = tf.constant(0, dtype=tf.float32)
        int_zero = tf.constant(0, dtype=tf.int64)
        trg_eos = tf.constant(self._trg_data_pipeline.meta["eos_id"], dtype=tf.int64)

        dataset = ds.build(map_func=self.get_data_preprocess_fn(mode, ds.status, args),
                           map_output_dtypes=self.inputs_signature(mode)[0],
                           auto_shard=(mode == compat.ModeKeys.TRAIN),
                           shuffle=(mode == compat.ModeKeys.TRAIN))

        if mode == compat.ModeKeys.INFER:
            logging.info("Creating test dataset.")
            return dataset.cache().padded_batch(
                dataset_utils.adjust_batch_size(args["batch_size"],
                                                num_replicas_in_sync=num_replicas_in_sync),
                padded_shapes={"audio": [None], "audio_length": []},
                padding_values={"audio": float_zero, "audio_length": int_zero},
                drop_remainder=False)

        elif mode == compat.ModeKeys.EVAL:
            logging.info("Creating evaluation dataset.")
            return dataset.cache().padded_batch(
                dataset_utils.adjust_batch_size(args["batch_size"],
                                                num_replicas_in_sync=num_replicas_in_sync),
                padded_shapes={"audio": [None], "audio_length": [], "transcript": [None]},
                padding_values={"audio": float_zero, "audio_length": int_zero, "transcript": trg_eos},
                drop_remainder=False)
        else:
            logging.info("Creating training dataset.")
            dataset = dataset_utils.clean_dataset_by_length(
                dataset, {"audio": args["max_src_len"] * self._audio_feature_dim * self._audio_feature_channels,
                          "audio_length": -1, "transcript": args["max_trg_len"]})
            if args["cache_dataset"]:
                dataset = dataset.cache()
            if args["shuffle_buffer"]:
                dataset = dataset.shuffle(buffer_size=args["shuffle_buffer"])
            padding_values = {"audio": float_zero, "audio_length": int_zero, "transcript": trg_eos}
            if args["max_src_len"] is None:
                raise RuntimeError("`max_src_len` for SpeechToText task must be provided.")
            if args["max_trg_len"] is None:
                raise RuntimeError("`max_trg_len` for SpeechToText task must be provided.")
            max_src_len = args["max_src_len"]
            max_trg_len = minimal_multiple(args["max_trg_len"], 8)
            audio_bucket_boundaries = create_audio_bucket_boundaries(max_src_len, args["min_src_bucket_boundary"])
            audio_bucket_boundaries[-1] = minimal_multiple(audio_bucket_boundaries[-1], 8)
            batch_size = dataset_utils.adjust_batch_size(args["batch_size"], args["batch_size_per_gpu"],
                                                         num_replicas_in_sync=num_replicas_in_sync,
                                                         verbose=False)
            batch_size_per_gpu = batch_size // num_replicas_in_sync
            assert batch_size_per_gpu > max_src_len, (
                f"batch size per gpu({batch_size_per_gpu} must be greater than "
                f"`max_src_len`={max_src_len}")
            if args["disable_batch_efficiency"]:
                bucket_batch_sizes = [int(batch_size_per_gpu // bound
                                          * num_replicas_in_sync) for bound in audio_bucket_boundaries]
            else:
                bucket_batch_sizes = [int(minimal_multiple(batch_size_per_gpu // bound, 8)
                                          * num_replicas_in_sync) for bound in audio_bucket_boundaries]
            frame_transcript_ratio = args["experimental_frame_transcript_ratio"]
            if frame_transcript_ratio is None:
                logging.info("WARNING: we recommend one to pre-scan the dataset and estimate the ratio: "
                             "frame length / transcript length.")
            else:
                trans_bucket_boundaries = [
                    int(bound / (frame_transcript_ratio + i * (
                        max_src_len / max_trg_len - frame_transcript_ratio) / len(audio_bucket_boundaries)))
                    for i, bound in enumerate(audio_bucket_boundaries)]
                trans_bucket_boundaries = [minimal_multiple(min(i, max_trg_len), 8) for i in
                                           trans_bucket_boundaries]
                num_buckets = len(trans_bucket_boundaries)
                true_trans_bucket_boundaries = []
                num_input_shapes = 0
                for idx, (batc, bound, tbound) in enumerate(zip(bucket_batch_sizes, audio_bucket_boundaries,
                                                                trans_bucket_boundaries)):
                    max_trans_len = [tbound,
                                     trans_bucket_boundaries[min(idx + 1, len(bucket_batch_sizes) - 1)]]
                    num_input_shapes += len(set(max_trans_len))
                    true_trans_bucket_boundaries.append(max_trans_len)
                logging.info(f"There are {num_input_shapes} input shapes to be compiled:")
                for idx, (batc, bound, tbound) in enumerate(zip(bucket_batch_sizes, audio_bucket_boundaries,
                                                                true_trans_bucket_boundaries)):
                    logging.info(f"   - batch={batc}, maximum-frames={bound}, "
                                 f"maximum-transcript-length={set(tbound)}")
                true_trans_bucket_boundaries = tf.constant(true_trans_bucket_boundaries, dtype=tf.int32)
                true_audio_bucket_boundaries = tf.transpose(tf.constant([audio_bucket_boundaries] * 2, dtype=tf.int32))

            bucket_batch_sizes = tf.constant(bucket_batch_sizes, dtype=tf.int64)
            audio_bucket_boundaries = tf.constant(audio_bucket_boundaries, dtype=tf.int32)

            def example_to_bucket_id(examples):
                """Return int64 bucket id for this example, calculated based on length."""
                if frame_transcript_ratio is None:
                    conditions_c = tf.less_equal(tf.cast(examples["audio_length"], tf.int32),
                                                 audio_bucket_boundaries)
                    return tf.reduce_min(tf.where(conditions_c))
                conditions_c = tf.logical_and(
                    tf.less_equal(tf.cast(examples["audio_length"], tf.int32), true_audio_bucket_boundaries),
                    tf.less_equal(tf.size(examples["transcript"]), true_trans_bucket_boundaries))
                minimum_match = tf.where(conditions_c)[0]
                return minimum_match[0] * num_buckets + minimum_match[1]

            def window_size_fn(bucket_id):
                """Return number of examples to be grouped when given a bucket id."""
                if frame_transcript_ratio is None:
                    return bucket_batch_sizes[bucket_id]
                return bucket_batch_sizes[bucket_id // num_buckets]

            def batching_fn(bucket_id, grouped_dataset):
                """Batch and add padding to a dataset of elements with similar lengths."""
                bucket_batch_size = window_size_fn(bucket_id)

                # Batch the dataset and add padding so that all input sequences in the
                # examples have the same length, and all target sequences have the same
                # lengths as well. Resulting lengths of inputs and targets can differ.
                return grouped_dataset.padded_batch(
                    bucket_batch_size,
                    padded_shapes={
                        "audio": ([(audio_bucket_boundaries[bucket_id] if frame_transcript_ratio is None
                                    else audio_bucket_boundaries[bucket_id // num_buckets])
                                   * self._audio_feature_dim * self._audio_feature_channels]),
                        "audio_length": [],
                        "transcript": ([None] if frame_transcript_ratio is None
                                       else [
                            true_trans_bucket_boundaries[bucket_id // num_buckets][bucket_id % num_buckets]])
                    },
                    padding_values=padding_values, drop_remainder=True)

            return dataset.apply(tf.data.experimental.group_by_window(
                key_func=example_to_bucket_id,
                reduce_func=batching_fn,
                window_size=None,
                window_size_func=window_size_fn))

    def build_metric_layer(self):
        return [AudioFramesMetricLayer("src"), SequenceTokenMetricLayer("trg")]

    def get_eval_metric(self, args, name="metric"):
        """ Returns a neurst.metrics.metric.Metric object for evaluation."""
        return build_metric(args[name + ".class"], language=self._trg_data_pipeline.meta["language"],
                            **args[name + ".params"])


@register_task
class MultiTaskSpeechTranslation(Task):

    def __init__(self, args):
        """ Initializes with configuration. """
        super(MultiTaskSpeechTranslation, self).__init__(args)
        transcript_dp_cls = args.get("transcript_data_pipeline.class", TranscriptDataPipeline)
        transcript_dp_params = args.get("transcript_data_pipeline.params", None) or {}
        self._transcript_data_pipeline = build_data_pipeline(
            transcript_dp_cls, **transcript_dp_params)
        translation_dp_cls = args.get("translation_data_pipeline.class", TranscriptDataPipeline)
        translation_dp_params = args.get("translation_data_pipeline.params", None) or {}
        self._translation_data_pipeline = build_data_pipeline(
            translation_dp_cls, **translation_dp_params)

    def get_config(self):
        return {
            "transcript_data_pipeline.class": self._transcript_data_pipeline.__class__.__name__,
            "transcript_data_pipeline.params": self._transcript_data_pipeline.get_config(),
            "translation_data_pipeline.class": self._translation_data_pipeline.__class__.__name__,
            "translation_data_pipeline.params": self._translation_data_pipeline.get_config(),
        }

    @staticmethod
    def class_or_method_args():
        """ Returns a list of args for flag definition. """
        this_args = super(SpeechToText, SpeechToText).class_or_method_args()
        this_args.extend([
            ModuleFlag("transcript_data_pipeline", DataPipeline.REGISTRY_NAME,
                       default=TranscriptDataPipeline.__name__,
                       help="The data pipeline for ASR transcription."),
            ModuleFlag("translation_data_pipeline", DataPipeline.REGISTRY_NAME,
                       default=TranscriptDataPipeline.__name__,
                       help="The data pipeline for translation."),
        ])
        return this_args

    def inputs_signature(self, mode) -> Tuple[dict, dict]:
        """ Returns the input dtypes and signatures (from dataset). """
        dtypes = {"audio": tf.float32, "audio_length": tf.int64}
        # [batch, frames, feature_dim]
        signatures = {"audio": tf.TensorShape([None, None]),
                      "audio_length": tf.TensorShape([None, ])}
        if mode == compat.ModeKeys.INFER:
            return dtypes, signatures
        dtypes["transcript"] = tf.int64
        signatures["transcript"] = tf.TensorShape([None, None])
        dtypes["translation"] = tf.int64
        signatures["translation"] = tf.TensorShape([None, None])
        return dtypes, signatures

    def example_to_input(self, batch_of_data: dict, mode) -> dict:
        """ Converts a batch of data into the model readable structure. """
        raise NotImplementedError

    def get_data_preprocess_fn(self, mode, data_status, args=None) -> callable:
        """ Returns a callable function that preprocess the data sample
            according to this task. """
        _ = args
        if mode != compat.ModeKeys.TRAIN:
            raise NotImplementedError

        # if data_status["audio"] != compat.DataStatus.PROJECTED:
        #     raise RuntimeError("We recommend one to preprocess the audio in advance.")

        def _process_text(text, status, dp):
            if status == compat.DataStatus.RAW:
                text = dp.process(text, is_processed=False)
            return text

        def _process(example):
            return {"audio": example["audio"],
                    "transcript": _process_text(example["transcript"], data_status["transcript"],
                                                self._transcript_data_pipeline),
                    "translation": _process_text(example["translation"], data_status["translation"],
                                                 self._translation_data_pipeline)
                    }

        return _process

    def get_data_postprocess_fn(self, mode) -> callable:
        """ Returns a callable function that postprocess the data sample
            according to this task. """
        raise NotImplementedError

    def create_and_batch_tfds(self,
                              ds: Dataset,
                              mode,
                              args=None,
                              num_replicas_in_sync=1) -> tf.data.Dataset:
        """ Batch dataset. """
        raise NotImplementedError

    def build_model(self, args, name=None):
        """Build a new model instance."""
        raise NotImplementedError



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import os
import re

import tensorflow as tf
import yaml
from absl import logging

from neurst.data.datasets import register_dataset
from neurst.data.datasets.audio.audio_dataset import RawAudioDataset
from neurst.utils.compat import DataStatus
from neurst.utils.flags_core import Flag

try:
    import soundfile
except (ImportError, OSError):
    pass


@register_dataset
class MuSTC(RawAudioDataset):
    """
    MuST-C(Gangi et al.,2019) is a multilingual speech translation corpus from English
    to 8 languages: Dutch (NL), French (FR), German (DE), Italian (IT), Portuguese (PT),
    Romanian (RO), Russian (RU) and Spanish (ES). MuST-C comprises at least 385 hours of
    audio recordings from English TED talks with their manual transcriptions and
    translations at sentence level for training.

    Homepage: https://ict.fbk.eu/must-c/
    """
    EXTRACTION_CHOICES = ["train", "dev", "tst", "test", "tst-HE", "tst-COMMON"]
    TARGET_LANGUAGES = ["nl", "fr", "de", "it", "pt", "ro", "ru", "es"]

    def __init__(self, args):
        super(MuSTC, self).__init__(args)
        self._extraction = args["extraction"]
        if self._extraction not in MuSTC.EXTRACTION_CHOICES:
            raise ValueError("`extraction` for MuST-C dataset must be "
                             "one of {}".format(", ".join(MuSTC.EXTRACTION_CHOICES)))
        if self._extraction in ["tst", "test"]:
            self._extraction = "tst-COMMON"
        self._transc_transla_dict = None

    @staticmethod
    def class_or_method_args():
        this_args = super(MuSTC, MuSTC).class_or_method_args()
        this_args.append(
            Flag("extraction", dtype=Flag.TYPE.STRING, default=None,
                 choices=MuSTC.EXTRACTION_CHOICES,
                 help="The dataset portion to be extracted, e.g. train, dev, test (tst-COMMON)."))
        return this_args

    @property
    def status(self):
        return {
            "audio": DataStatus.RAW,
            "transcript": DataStatus.RAW,
            "translation": DataStatus.RAW
        }

    def load_transcripts(self):
        """ Loads transcripts and translations. """
        if self._transc_transla_dict is not None:
            return
        logging.info(f"Loading transcriptions from tarball: {self._input_tarball}")
        srcs = None
        trgs = None
        segments = None
        if tf.io.gfile.isdir(self._input_tarball):
            path = os.path.join(self._input_tarball, f"data/{self._extraction}/txt")
            for filename in tf.io.gfile.listdir(path):
                with tf.io.gfile.GFile(os.path.join(path, filename), "rb") as fp:
                    if filename.endswith(".yaml"):
                        segments = yaml.load(fp.read().decode("utf-8"), Loader=yaml.FullLoader)
                    elif filename.endswith(".en"):
                        srcs = fp.read().decode("utf-8").strip().split("\n")
                    elif filename[-2:] in MuSTC.TARGET_LANGUAGES:
                        trgs = fp.read().decode("utf-8").strip().split("\n")
                    if srcs is not None and trgs is not None and segments is not None:
                        break
        else:
            with self.open_tarball("tar") as tar:
                for tarinfo in tar:
                    if not tarinfo.isreg():
                        continue
                    if tarinfo.name.endswith(f"{self._extraction}.yaml"):
                        f = tar.extractfile(tarinfo)
                        segments = yaml.load(f.read().decode("utf-8"), Loader=yaml.FullLoader)
                        f.close()
                    elif tarinfo.name.endswith(f"{self._extraction}.en"):
                        f = tar.extractfile(tarinfo)
                        srcs = f.read().decode("utf-8").strip().split("\n")
                        f.close()
                    elif re.match(r"^.*/{}.({})$".format(self._extraction, "|".join(MuSTC.TARGET_LANGUAGES)),
                                  tarinfo.name):
                        f = tar.extractfile(tarinfo)
                        trgs = f.read().decode("utf-8").strip().split("\n")
                        f.close()
                    if srcs is not None and trgs is not None and segments is not None:
                        break
        assert len(srcs) == len(trgs) == len(segments)
        self._transc_transla_dict = {}
        self._transcripts = []
        self._translations = []
        n = 0
        for src, trg, seg in zip(srcs, trgs, segments):
            src = src.strip()
            trg = trg.strip()
            if not src or not trg:
                continue
            if seg["wav"] not in self._transc_transla_dict:
                self._transc_transla_dict[seg["wav"]] = []
            self._transc_transla_dict[seg["wav"]].append([float(seg["offset"]),
                                                          float(seg["duration"]), src, trg])
            n += 1
            self._transcripts.append(src)
            self._translations.append(trg)
        logging.info("Total %d utterances (%d skipped).", n, len(srcs) - n)

    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
        """
        if total_shards > 1:
            total_samples = self.num_samples
            samples_per_part = total_samples // total_shards
            range_begin = samples_per_part * shard_id
            if shard_id == total_shards - 1:
                range_end = total_samples + 1
                logging.info(f"Iterate on dataset from {range_begin} "
                             f"to the end (total {total_samples}).")
            else:
                range_end = range_begin + samples_per_part
                logging.info(f"Iterate on dataset from {range_begin} "
                             f"to {range_end} (total {total_samples}).")

        def gen():
            if self._transc_transla_dict is None:
                self.load_transcripts()
            n = 0
            former_wavkey = None
            hit_end = False
            ori_audio = None
            sample_rate = None
            num_processed = 0
            if tf.io.gfile.isdir(self._input_tarball):
                for wavname in self._transc_transla_dict:
                    for offset, duration, transcript, transla in self._transc_transla_dict[wavname]:
                        n += 1
                        if total_shards > 1:
                            if n < range_begin:
                                continue
                            if n >= range_end:
                                hit_end = True
                                break
                        if wavname != former_wavkey:
                            former_wavkey = wavname
                            ori_audio, sample_rate = soundfile.read(
                                os.path.join(self._input_tarball, f"data/{self._extraction}/wav/{wavname}"),
                                dtype='int16')
                        start = int(offset * sample_rate)
                        end = int((offset + duration) * sample_rate) + 1
                        data_sample = {
                            "audio": self.extract_audio_feature(
                                sig=ori_audio[start:end], rate=sample_rate),
                            "transcript": transcript,
                            "translation": transla,
                        }
                        if map_func is None:
                            yield data_sample
                        else:
                            yield map_func(data_sample)
                    if hit_end:
                        break

            else:
                with self.open_tarball("tar") as tar:
                    for tarinfo in tar:
                        if num_processed == len(self._transc_transla_dict):
                            break
                        if not tarinfo.isreg():
                            continue
                        if self._extraction not in tarinfo.name:
                            continue
                        if not tarinfo.name.endswith(".wav"):
                            continue
                        wavname = tarinfo.name.split("/")[-1]
                        if wavname not in self._transc_transla_dict:
                            continue
                        num_processed += 1
                        for offset, duration, transcript, transla in self._transc_transla_dict[wavname]:
                            n += 1
                            if total_shards > 1:
                                if n < range_begin:
                                    continue
                                if n >= range_end:
                                    hit_end = True
                                    break
                            if wavname != former_wavkey:
                                former_wavkey = wavname
                                f = tar.extractfile(tarinfo)
                                b = io.BytesIO(f.read())
                                f.close()
                                ori_audio, sample_rate = soundfile.read(b, dtype='int16')
                                b.close()
                            start = int(offset * sample_rate)
                            end = int((offset + duration) * sample_rate) + 1
                            data_sample = {
                                "audio": self.extract_audio_feature(
                                    sig=ori_audio[start:end], rate=sample_rate),
                                "transcript": transcript,
                                "translation": transla,
                            }
                            if map_func is None:
                                yield data_sample
                            else:
                                yield map_func(data_sample)
                        if hit_end:
                            break

        return gen



import numpy
import tensorflow as tf

from neurst.layers.layer_utils import one_entry_bias, stack_beam_size, tile_tensor
from neurst.utils import compat


def tf1codebase_stack_beam_size(tensors, beam_size):
    """ Stacks the tensors `beam_size` times at specific dimension.

    Args:
        tensors: A Tensor of a list/tuple/dict of Tensors. For each Tensor, the first
          dimension must be batch_size, otherwise, unknow errors may occur.
        beam_size: A python integer, the beam width.

    Returns: A Tensor or a list/tuple of Tensors with the same structure
      as `tensors`.

    Raises:
        AssertionError: if the shape of tensor does not match
          [batch_size, 1, 1, timesteps] when tensor.ndims == 4.
        NotImplementedError: if tensor.ndims > 4.
    """

    def _stack(x):
        assert isinstance(x, tf.Tensor)
        batch_size = tf.shape(x)[0]
        x_ndims = x.get_shape().ndims
        last_dim = x.get_shape().as_list()[-1]
        if last_dim is None:
            last_dim = tf.shape(x)[-1]
        if x_ndims == 3:
            final_shape = [beam_size * batch_size, -1, last_dim]
            stacked_x = tf.reshape(tf.tile(x, [1, beam_size, 1]), final_shape)
            return stacked_x
        elif x_ndims == 2:
            final_shape = [batch_size * beam_size, last_dim]
            return tf.reshape(tf.tile(x, [1, beam_size]), final_shape)
        elif x_ndims == 1:
            return tf.reshape(
                tf.transpose(tf.tile([x], [beam_size, 1])), [-1])
        elif x_ndims == 4:
            assert x.get_shape().as_list()[1] == x.get_shape().as_list()[2] == 1, (
                "this only matches the bias tensor with shape [batch_size, 1, 1, timesteps]")
            return tf.expand_dims(
                _stack(tf.squeeze(x, axis=1)), axis=1)
        else:
            raise NotImplementedError("Not implemented the capability for ndims={}".format(x_ndims))

    return tf.nest.pack_sequence_as(
        tensors,
        tf.nest.map_structure(
            _stack, tf.nest.flatten(tensors)))


def tf1codebase_finished_beam_one_entry_bias(on_entry, num_entries, dtype=None):
    """ Builds a bias vector to be added to log_probs of a finished beam.

    The returned vector with shape [`num_entries`, ]. Only the `on_entry`
    has value 0, and the others are FLOAT_MIN.

    For example, on_entry=3 and num_entries=6 get the vector
    [FLOAT_MIN, FLOAT_MIN, FLOAT_MIN, 0.0, FLOAT_MIN, FLOAT_MIN]

    Args:
        on_entry: A python integer.
        num_entries: A python integer.
        dtype:

    Returns: A bias vector.
    """
    if dtype is None:
        dtype = compat.CUSTOM_GLOBAL_FLOATX
    bias = tf.one_hot(
        [on_entry], num_entries,
        on_value=0., off_value=compat.FLOAT_MIN)
    if dtype.base_dtype.name != "float32":
        bias = tf.cast(bias, dtype=dtype)
    return tf.squeeze(bias, axis=0)


def tf1codebase_unk_beam_one_entry_bias(on_entry, num_entries, dtype=None):
    """ Builds a bias vector to be added to log_probs of the UNK token in the vocab.

    The returned vector with shape [`num_entries`, ]. Only the `on_entry`
    has value FLOAT_MIN, and the others are 0.

    For example, on_entry=3 and num_entries=6 get the vector
    [FLOAT_MIN, FLOAT_MIN, FLOAT_MIN, 0.0, FLOAT_MIN, FLOAT_MIN]

    Args:
        on_entry: A python integer.
        num_entries: A python integer.
        dtype:

    Returns: A bias vector.
    """
    if dtype is None:
        dtype = compat.CUSTOM_GLOBAL_FLOATX
    bias = tf.one_hot(
        [on_entry], num_entries,
        on_value=compat.FLOAT_MIN, off_value=0.)
    if dtype.base_dtype.name != "float32":
        bias = tf.cast(bias, dtype=dtype)
    return tf.squeeze(bias, axis=0)


def tf1codebase_expand_to_beam_size(tensor, beam_size, axis=0):
    """ Stacks a given tensor `beam_size` times on a specific axis.

    For example, tensor=[1, 2, 3, 4], beam_size=3, axis=0 get the tensor
    [ [1, 2, 3, 4],
      [1, 2, 3, 4],
      [1, 2, 3, 4] ]

    tensor=[[1, 2, 3], [3, 4, 5]], beam_size=1, axis=1 get the tensor
    [ [[1, 2, 3]], [[3, 4, 5]] ]

    Args:
        tensor: A Tensor.
        beam_size: A python integer, the beam width.
        axis: A python integer.

    Returns: A Tensor.
    """
    tensor = tf.expand_dims(tensor, axis=axis)
    tile_dims = [1] * tensor.get_shape().ndims
    tile_dims[axis] = beam_size
    return tf.tile(tensor, tile_dims)


def test_fn_stack_beam_size():
    beam_size = 4
    batch_size = 3
    dim = 5
    time_step = 7
    num_heads = 8
    tensor_1d = tf.convert_to_tensor(numpy.random.randint(0, 10, (batch_size,)), dtype=tf.int32)
    tensor_2d = tf.convert_to_tensor(numpy.random.rand(batch_size, dim), dtype=tf.float32)
    tensor_3d = tf.convert_to_tensor(numpy.random.rand(batch_size, time_step, dim), dtype=tf.float32)
    tensor_4d_bias = tf.convert_to_tensor(numpy.random.rand(batch_size, 1, 1, dim), dtype=tf.float32)
    tensor_4d = tf.convert_to_tensor(numpy.random.rand(batch_size, time_step, num_heads, dim), dtype=tf.float32)
    tensor_3d_from_4d = tf.reshape(tensor_4d, [batch_size, time_step, -1])
    assert (stack_beam_size(tensor_1d, beam_size).numpy()
            == tf1codebase_stack_beam_size(tensor_1d, beam_size).numpy()).all()
    assert (stack_beam_size(tensor_2d, beam_size).numpy()
            == tf1codebase_stack_beam_size(tensor_2d, beam_size).numpy()).all()
    assert (stack_beam_size(tensor_3d, beam_size).numpy()
            == tf1codebase_stack_beam_size(tensor_3d, beam_size).numpy()).all()
    assert (stack_beam_size(tensor_4d_bias, beam_size).numpy()
            == tf1codebase_stack_beam_size(tensor_4d_bias, beam_size).numpy()).all()
    assert (stack_beam_size(tensor_4d, beam_size).numpy()
            == tf.reshape(tf1codebase_stack_beam_size(tensor_3d_from_4d, beam_size),
                          [-1, time_step, num_heads, dim]).numpy()).all()


def test_fn_one_entry_bias():
    vocab_size = 10
    eos_id = 9
    unk_id = 7
    assert (tf1codebase_finished_beam_one_entry_bias(on_entry=eos_id,
                                                     num_entries=vocab_size, dtype=tf.float32).numpy()
            == one_entry_bias(on_entry=eos_id, num_entries=vocab_size, on_value=0.,
                              off_value=compat.FLOAT_MIN, dtype=tf.float32).numpy()).all()
    assert (tf1codebase_unk_beam_one_entry_bias(on_entry=unk_id,
                                                num_entries=vocab_size, dtype=tf.float32).numpy()
            == one_entry_bias(on_entry=unk_id, num_entries=vocab_size, off_value=0.,
                              on_value=compat.FLOAT_MIN, dtype=tf.float32).numpy()).all()


def test_fn_expand_tensor():
    vocab_size = 10
    eos_id = 9
    batch_size = 3
    beam_size = 4
    batch_beam_size = batch_size * beam_size
    finished_beam_bias = tf1codebase_finished_beam_one_entry_bias(
        on_entry=eos_id, num_entries=vocab_size, dtype=tf.float32)
    assert (tf1codebase_expand_to_beam_size(finished_beam_bias,
                                            batch_beam_size, axis=0).numpy()
            == tile_tensor(finished_beam_bias,
                           batch_beam_size, axis=0).numpy()).all()


if __name__ == "__main__":
    test_fn_stack_beam_size()
    test_fn_one_entry_bias()
    test_fn_expand_tensor()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import logging

from neurst.optimizers.schedules import register_lr_schedule
from neurst.utils import compat
from neurst.utils.flags_core import Flag


@register_lr_schedule("piecewise")
class PiecewiseSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Learning rate schedule."""

    def __init__(self, args):
        """Initialize configuration of the learning rate schedule.

        Args:
          args: A dict of full parameters.
        """
        super(PiecewiseSchedule, self).__init__()
        self._schedule_steps = args["schedule_steps"]
        self._schedule_lrs = args["schedule_lrs"]
        assert len(self._schedule_steps) + 1 == len(self._schedule_lrs)
        self._initial_step = compat.get_registered_initial_step()
        logging.info("Initialize PiecewiseSchedule from global step={}. "
                     "The learning rate will be:".format(self._initial_step))
        for idx, (step, lr) in enumerate(zip(self._schedule_steps, self._schedule_lrs)):
            if idx == 0:
                logging.info("    linear warmup from 0~{} for {} steps".format(lr, step))
            else:
                logging.info("    {} from step={} to step={}".format(lr, self._schedule_steps[idx - 1], step))
        logging.info("    {} for step>{}".format(self._schedule_lrs[-1], self._schedule_steps[-1]))

    @staticmethod
    def class_or_method_args():
        return [
            Flag("schedule_steps", dtype=Flag.TYPE.STRING, default=None,
                 help="A list of triggered steps."),
            Flag("schedule_lrs", dtype=Flag.TYPE.STRING, default=None,
                 help="A list of learning rates."),
        ]

    def __call__(self, global_step):
        """Calculate learning rate.

        Args:
          global_step: An integer, the current global step used for learning rate
            calculation.

        Returns:
          A float, the learning rate needs to be used for current global step.
        """
        with tf.name_scope('learning_rate_schedule'):
            global_step = tf.cast(global_step, tf.float32) + self._initial_step + 1.
            pred_fn_pairs = [(tf.less(global_step, self._schedule_steps[0]),
                              lambda: self._schedule_lrs[0] / float(self._schedule_steps[0]) * global_step)]
            for step, lr in zip(self._schedule_steps[1:], self._schedule_lrs[1:-1]):
                pred_fn_pairs.append((tf.less(global_step, step), lambda: tf.constant(lr)))

            return tf.case(pred_fn_pairs, default=(lambda: tf.constant(self._schedule_lrs[-1])))

    def get_config(self):
        """Get the configuration of the learning rate schedule."""
        return {
            "schedule_steps": self._schedule_steps,
            "schedule_lrs": self._schedule_lrs
        }



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" neurst only"""
import numpy
import tensorflow as tf

from neurst.layers.encoders.transformer_encoder import TransformerEncoder
from neurst.layers.modalities.audio_modalities import PositionalConv, Wav2vec2FeatureExtractor
from neurst.models import register_model
from neurst.models.model import BaseModel
from neurst.utils import compat
from neurst.utils.flags_core import Flag
from neurst.utils.hparams_sets import register_hparams_set


@register_model("wav2vec2")
class Wav2Vec2(BaseModel):
    """ Defines the Wav2vec2 model.
    Described in https://arxiv.org/abs/2006.11477.
    """

    def __init__(self,
                 args: dict,
                 feature_extractor,
                 post_extract_proj,
                 positional_conv,
                 encoder,
                 name=None):
        """ Initializes a BERT model.

        Args:
            args: A dict, containing the model configuration.
            feature_extractor: The convolutional layers for raw audio feature extraction.
            post_extract_proj: A dense layer or None.
            positional_conv: The positional convolution layer.
            encoder: The transformer encoder for feature transformation.
            name: The name of the model.options = tf.data.Options()
        """
        super(Wav2Vec2, self).__init__(args, name=name or "wav2vec2")
        self._feature_extractor = feature_extractor
        self._feature_extractor_norm_layer = tf.keras.layers.LayerNormalization(
            epsilon=1.e-5, dtype="float32", name="feature_extractor_norm")
        self._post_extract_proj = post_extract_proj
        self._pos_conv = positional_conv
        self._encoder = encoder
        self._dropout_input = args["dropout_input"]
        self._encoder_dropout = args["encoder_dropout"]

    @staticmethod
    def class_or_method_args():
        return [
            Flag("conv_bias", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to use bias tensor in feature extractor."),
            Flag("conv_feature_layers", dtype=Flag.TYPE.STRING,
                 default="[(512, 10, 5)] + [(512, 3, 2)] * 4 + [(512,2,2)] + [(512,2,2)]",
                 help="The string describing convolutional feature extraction layers "
                      "in form of a python list that contains [(dim, kernel_size, stride), ...]"),
            Flag("extractor_mode", dtype=Flag.TYPE.STRING, default="default",
                 choices=["default", "layer_norm"],
                 help="The mode for feature extractor. default has a single group norm with d "
                      "groups in the first conv block, whereas layer_norm has layer norms in "
                      "every block (meant to use with normalize=True)"),
            Flag("encoder_embed_dim", dtype=Flag.TYPE.INTEGER, default=768,
                 help="The dimension of embedding."),
            Flag("conv_pos", dtype=Flag.TYPE.INTEGER, default=128,
                 help="The number of filters for convolutional positional embeddings."),
            Flag("conv_pos_groups", dtype=Flag.TYPE.INTEGER, default=16,
                 help="The number of groups for convolutional positional embedding."),
            Flag("dropout_input", dtype=Flag.TYPE.FLOAT, default=0.1,
                 help="The dropout rate to apply to the input (after feat extr)."),
            Flag("encoder_layerdrop", dtype=Flag.TYPE.FLOAT, default=0.05,
                 help="The probability of dropping a transformer layer."),
            Flag("encoder_dropout", dtype=Flag.TYPE.FLOAT, default=0.1,
                 help="The dropout probability for the transformer encoder."),
            Flag("encoder_layers", dtype=Flag.TYPE.INTEGER, default=12,
                 help="The number of layers in the transformer encoder."),
            Flag("encoder_attention_heads", dtype=Flag.TYPE.INTEGER, default=12,
                 help="The number of attention heads in the transformer encoder."),
            Flag("encoder_filter_size", dtype=Flag.TYPE.INTEGER, default=3072,
                 help="The number of units of the ffn layer in the transformer encoder."),
            Flag("encoder_activation_fn", dtype=Flag.TYPE.STRING, default="gelu_nonapprox",
                 help="The activation function to be used in the ffn layer of transformer encoder."),
            Flag("encoder_pre_norm", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="The norm mode for the transformer encoder."),
            Flag("quantize_input", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to use quantized inputs.")
        ]

    @classmethod
    def new(cls, args: dict, name=None):
        """ Builds a sequence to sequence model.

        Args:
            args: A dict containing all model parameters.
            name: The name of the model.

        Returns:
            A Wav2Vec2 model.
        """
        args["conv_feature_layers"] = eval(args["conv_feature_layers"])
        feature_extractor = Wav2vec2FeatureExtractor(conv_layers=args["conv_feature_layers"],
                                                     mode=args["extractor_mode"],
                                                     conv_bias=args["conv_bias"],
                                                     dropout=0., name="feature_extractor")
        post_extract_proj = None
        if (args["conv_feature_layers"][-1][0] != args["encoder_embed_dim"]
            and not args["quantize_input"]):
            post_extract_proj = tf.keras.layers.Dense(args["encoder_embed_dim"], name="post_extract_proj")
        pos_conv = PositionalConv(args["encoder_embed_dim"], kernel_size=args["conv_pos"],
                                  groups=args["conv_pos_groups"], dropout=args["encoder_dropout"], name="pos_conv")
        encoder = TransformerEncoder(num_layers=args["encoder_layers"],
                                     hidden_size=args["encoder_embed_dim"],
                                     num_attention_heads=args["encoder_attention_heads"],
                                     filter_size=args["encoder_filter_size"],
                                     ffn_activation=args["encoder_activation_fn"],
                                     attention_dropout_rate=args["encoder_dropout"],
                                     ffn_dropout_rate=args["encoder_dropout"],
                                     layer_postprocess_dropout_rate=args["encoder_dropout"],
                                     post_normalize=(not args["encoder_pre_norm"]),
                                     layer_postprocess_epsilon=1e-5, name="encoder")
        model = cls(args, feature_extractor, post_extract_proj, pos_conv, encoder)
        _src = tf.convert_to_tensor(numpy.random.rand(1, 5000), dtype=compat.CUSTOM_GLOBAL_FLOATX)
        _ = model({"src": _src, "src_padding": tf.zeros_like(_src)})
        return model

    def call(self, inputs, is_training=True):
        """ Forward pass of the BERT classifier.

        Args:
            inputs: A dict containing model inputs.
                - src: The raw audio signals with shape [batch, max_signals]
                - src_padding: A bool tensor indicating padding or not, with shape [batch, max_signals]
            is_training:

        Returns: A dict containing

        """
        src = inputs["src"]
        src_padding = inputs.get("src_padding", None)
        features = self._feature_extractor(src, is_training=is_training)
        # features_pen = tf.reduce_mean(tf.math.pow(features, 2))
        features = tf.cast(self._feature_extractor_norm_layer(features), compat.CUSTOM_GLOBAL_FLOATX)
        if src_padding is not None:
            ori_signal_size = tf.shape(src_padding)[1]
            feature_frames = tf.shape(features)[1]
            extra = ori_signal_size // feature_frames
            src_padding = tf.cast(tf.reduce_all(tf.cast(
                tf.reshape(src_padding[:, :extra * feature_frames],
                           [tf.shape(features)[0], feature_frames, -1]), tf.bool), axis=-1), features.dtype)

        if self._post_extract_proj is not None:
            features = self._post_extract_proj(features)
        if is_training:
            features = tf.nn.dropout(features, rate=self._dropout_input)
            # TODO: time mask
        x = self._pos_conv(features, src_padding, is_training=is_training)
        if is_training:
            x = tf.nn.dropout(x, rate=self._encoder_dropout)
        # TODO layer drop for training
        y = self._encoder(x, src_padding, is_training=False)
        return {
            "contextualized_representation": y,
            "contextualized_representation_padding": src_padding
        }


@register_hparams_set("wav2vec2_base")
def wav2vec2_base():
    return {
        "model.class": Wav2Vec2.__name__,
        "model.params": {
            "conv_bias": False,
            "conv_feature_layers": "[(512, 10, 5)] + [(512, 3, 2)] * 4 + [(512,2,2)] + [(512,2,2)]",
            "extractor_mode": "default",
            "encoder_embed_dim": 768,
            "conv_pos": 128,
            "quantize_input": False,
            "conv_pos_groups": 16,
            "dropout_input": 0.1,
            "encoder_layerdrop": 0.05,
            "encoder_dropout": 0.1,
            "encoder_layers": 12,
            "encoder_attention_heads": 12,
            "encoder_filter_size": 3072,
            "encoder_activation_fn": "gelu_nonapprox",
            "encoder_pre_norm": False
        }
    }


@register_hparams_set("wav2vec2_large")
def wav2vec2_large():
    return {
        "model.class": Wav2Vec2.__name__,
        "model.params": {
            "conv_bias": False,
            "conv_feature_layers": "[(512, 10, 5)] + [(512, 3, 2)] * 4 + [(512,2,2)] + [(512,2,2)]",
            "extractor_mode": "default",
            "encoder_embed_dim": 1024,
            "conv_pos": 128,
            "quantize_input": False,
            "conv_pos_groups": 16,
            "dropout_input": 0.1,
            "encoder_layerdrop": 0.2,
            "encoder_dropout": 0.1,
            "encoder_layers": 24,
            "encoder_attention_heads": 16,
            "encoder_filter_size": 4096,
            "encoder_activation_fn": "gelu_nonapprox",
            "encoder_pre_norm": False
        }
    }


@register_hparams_set("wav2vec2_toy")
def wav2vec2_toy():
    return {
        "model.class": Wav2Vec2.__name__,
        "model.params": {
            "conv_bias": False,
            "conv_feature_layers": "[(16, 10, 5)] + [(16, 3, 2)] * 4 + [(16,2,2)] + [(16,2,2)]",
            "extractor_mode": "default",
            "encoder_embed_dim": 16,
            "conv_pos": 2,
            "quantize_input": False,
            "conv_pos_groups": 1,
            "dropout_input": 0.1,
            "encoder_layerdrop": 0,
            "encoder_dropout": 0.1,
            "encoder_layers": 3,
            "encoder_attention_heads": 2,
            "encoder_filter_size": 32,
            "encoder_activation_fn": "gelu_nonapprox",
            "encoder_pre_norm": False
        }
    }



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import os

import tensorflow as tf
import yaml
from absl import logging


def copy_dict_list(d):
    """ Repeat a python build-in dict. """
    if type(d) in [int, float, bool, str]:
        return d
    try:
        new_d = {}
        for k, v in d.items():
            new_d[k] = copy_dict_list(v)
        return new_d
    except AttributeError:
        try:
            new_d = []
            for v in d:
                new_d.append(v)
            return new_d
        except TypeError:
            return d


def print_params(title, params, with_none=True, indent=0):
    """ Prints parameters.

    Args:
        title: A string.
        params: A dict.
    """
    if indent > 0:
        indent = "".join([" "] * indent)
    else:
        indent = ""

    def _params_to_stringlist(params, prefix):
        """ Convert a dictionary/list of parameters to a
            formatted string for logging purpose.

        Args:
            params: A dictionary/list of parameters.
            prefix: A string.

        Returns: A format string.

        Raises:
            ValueError: if unknown type of `params`.
        """
        param_list = []
        if isinstance(params, dict):
            for key, val in params.items():
                if not with_none and val is None:
                    continue
                param_list.append(prefix + key + ": ")
                if isinstance(val, dict) or isinstance(val, list):
                    param_list.extend(_params_to_stringlist(val, prefix + "  "))
                else:
                    param_list[-1] += str(val)
        elif isinstance(params, list):
            prefix += "  "
            for item in params:
                for idx, (key, val) in enumerate(item.items()):
                    if idx == 0:
                        newprefix = copy.deepcopy(prefix[:-2])
                        newprefix += "- "
                        param_list.append(newprefix + key + ": ")
                    else:
                        param_list.append(prefix + key + ": ")
                    if isinstance(val, dict):
                        param_list.extend(_params_to_stringlist(val, prefix + "  "))
                    else:
                        param_list[-1] += str(val)
        else:
            raise ValueError("Unrecognized type of params: {}".format(str(params)))
        return param_list

    logging.info(title)
    if params is None:
        return
    for info in _params_to_stringlist(params, "  " + indent):
        logging.info(info)


def extract_constructor_params(locals_of_this_fn,
                               keep_out_list=None,
                               verbose=True,
                               verbose_title=None):
    """ Removes some elements in `locals()` called in the beginning of constructor
        and does logging.

    Args:
        locals_of_this_fn: The `locals()` of a function.
        keep_out_list: A list of keepout keys.
        verbose: A bool, whether to do logging.
        verbose_title: A string, the logging title.

    Returns:
        A dict.
    """
    if keep_out_list is None:
        keep_out_list = []
    assert isinstance(keep_out_list, list), (
        "`keep_out_list` must be a list of strings or none.")
    params = dict()
    for k, v in locals_of_this_fn.items():
        if k in ["self", "_"] or k.startswith("__") or k in keep_out_list:
            continue
        elif k == "kwargs" and isinstance(v, dict):
            for kk, vv in v.items():
                params[kk] = vv
        else:
            params[k] = v
    if verbose:
        vb = params.pop("verbose", None)
        if vb is None or vb:
            if not verbose_title:

                try:
                    name = params.pop("name", None)
                    verbose_title = locals_of_this_fn["self"].__class__.__name__
                    if name:
                        verbose_title = name + "(class: {})".format(
                            verbose_title)
                    verbose_title = "Parameters for " + verbose_title
                except KeyError:
                    verbose_title = "Parameters for "
            print_params(title=verbose_title, params=params)
            if "name" in locals_of_this_fn:
                params["name"] = name
        if "verbose" in locals_of_this_fn:
            params["verbose"] = vb
    return params


def parse_params(params, default_params):
    """Parses parameter values to the types defined by the default parameters.
    Default parameters are used for missing values.

    Args:
        params: A dict.
        default_params: A dict to provide parameter structure and missing values.

    Returns: A updated dict.
    """
    # Cast parameters to correct types
    if params is None:
        params = {}
    result = copy.deepcopy(default_params)
    for key, value in params.items():
        # If param is unknown, drop it to stay compatible with past versions
        if key not in default_params:
            raise ValueError("{} is not a valid model parameter".format(key))
        # Param is a dictionary
        if isinstance(value, dict):
            default_dict = default_params[key]
            if not isinstance(default_dict, dict):
                raise ValueError("{} should not be a dictionary".format(key))
            if default_dict:
                value = parse_params(value, default_dict)
            else:
                # If the default is an empty dict we do not typecheck it
                # and assume it's done downstream
                pass
        if value is None:
            continue
        if default_params[key] is None:
            result[key] = value
        else:
            result[key] = type(default_params[key])(value)
    return result


def load_from_config_path(config_paths):
    """ Loads configurations from files of yaml format.

    Args:
        config_paths: A string (each file name is seperated by ",") or
          a list of strings (file names).

    Returns: A dictionary of model configurations, parsed from config files.
    """
    if config_paths is None:
        return {}
    if isinstance(config_paths, str):
        config_paths = config_paths.strip().split(",")
    assert isinstance(config_paths, list) or isinstance(config_paths, tuple)
    model_configs = dict()
    for config_path in config_paths:
        config_path = config_path.strip()
        if not config_path:
            continue
        if not tf.io.gfile.exists(config_path):
            raise OSError("config file does not exist: {config_path}".format(
                config_path=config_path))
        logging.info("loading configurations from {config_path}".format(
            config_path=config_path))
        with tf.io.gfile.GFile(config_path, "r") as config_file:
            config_flags = yaml.load(config_file, Loader=yaml.FullLoader)
            model_configs = deep_merge_dict(model_configs, config_flags)
    return model_configs


def deep_merge_dict(dict_x, dict_y, path=None, local_overwrite=True,
                    merge_only_exist=False, raise_exception=True):
    """ Recursively merges dict_y into dict_x.

    Args:
        dict_x: A dict.
        dict_y: A dict.
        path:
        local_overwrite: Whether to overwrite the `dict_x`.
        merge_only_exist: A bool, only keys in dict_x will be overwritten if True
            otherwise all key-value pairs in dict_y will be written into dict_x.
        raise_exception: A bool, whether to raise KerError exception when keys in
            dict_y but not in dict_x when merge_only_exist=True,

    Returns:
        An updated dict of dict_x

    Raises:
        KeyError: keys in dict_y but not in dict_x when merge_only_exist=True
            and raise_exception=True.
    """
    if path is None:
        path = []
    if not local_overwrite:
        dict_x = copy.deepcopy(dict_x)
    for key in dict_y:
        if dict_y[key] is None:
            if key not in dict_x:
                dict_x[key] = None
            continue
        if key in dict_x:
            if isinstance(dict_x[key], dict) and isinstance(dict_y[key], dict):
                deep_merge_dict(dict_x[key], dict_y[key], path + [str(key)],
                                local_overwrite=local_overwrite,
                                merge_only_exist=merge_only_exist)
            elif dict_x[key] == dict_y[key]:
                pass  # same leaf value
            else:
                dict_x[key] = dict_y[key]
        else:
            if merge_only_exist:
                if raise_exception:
                    raise KeyError("Key {} not in dict_x.".format(key))
            else:
                dict_x[key] = dict_y[key]
    return dict_x


class ModelConfigs:
    """ A class for dumping and loading model configurations. """

    MODEL_CONFIG_YAML_FILENAME = "model_configs.yml"

    @staticmethod
    def dump(model_config, output_dir, **kwaux):
        """ Dumps model configurations.

        Args:
            model_config: A dict.
            output_dir: A string, the output directory.
            kwaux: Arbitrary extra parameters.
        """
        cfg = copy.deepcopy(model_config)
        cfg = deep_merge_dict(cfg, kwaux, merge_only_exist=False)
        model_config_filename = os.path.join(output_dir, ModelConfigs.MODEL_CONFIG_YAML_FILENAME)
        if not tf.io.gfile.exists(output_dir):
            tf.io.gfile.makedirs(output_dir)
        logging.info("Saving model configurations to directory: {}".format(output_dir))
        with tf.io.gfile.GFile(model_config_filename, "w") as file:
            yaml.dump(cfg, file, default_flow_style=False)

    @staticmethod
    def load(model_dir):
        """ Loads model configurations.

        Args:
            model_dir: A string, the directory.

        Returns: A dict.
        """
        if not tf.io.gfile.isdir(model_dir):
            model_dir = os.path.dirname(model_dir)
        model_config_filename = os.path.join(model_dir, ModelConfigs.MODEL_CONFIG_YAML_FILENAME)
        if not tf.io.gfile.exists(model_config_filename):
            raise FileNotFoundError("Fail to find model config file: {}".format(model_config_filename))
        with tf.io.gfile.GFile(model_config_filename, "r") as file:
            model_configs = yaml.load(file, Loader=yaml.FullLoader)
        logging.info("Loading models configs from {model_config_filename}".format(
            model_config_filename=model_config_filename))
        return model_configs


def yaml_load_checking(args):
    for k in args:
        if args[k] and isinstance(args[k], str):
            try:
                v = yaml.load(args[k], Loader=yaml.FullLoader)
                args[k] = v
            except yaml.YAMLError:
                pass
    return args



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import os

from absl import logging

from neurst.data.datasets import register_dataset
from neurst.data.datasets.audio.audio_dataset import RawAudioDataset
from neurst.utils.compat import DataStatus


@register_dataset("AugmentedLibriSpeech")
class AugLibriSpeech(RawAudioDataset):
    """
    Augmented LibriSpeech is a small EN->FR dataset which
    was originally started from the LibriSpeech corpus.
    The English utterances were automatically aligned to
    the e-books in French and 236 hours of English speech
    aligned to French translations at utterance level were
    finally extracted. It has been widely used in previous studies.
    As such, we use the clean 100-hour portion plus the augmented
    machine translation from Google Translate as the training data.

    Homepage: https://github.com/alicank/Translation-Augmented-LibriSpeech-Corpus
    The raw dataset contains 3 files:
        - train_100h.zip
        - dev.zip
        - test.zip
    """

    def __init__(self, args):
        super(AugLibriSpeech, self).__init__(args)
        self._transc_transla_list = None

    @property
    def status(self):
        return {
            "audio": DataStatus.RAW,
            "transcript": DataStatus.RAW,
            "translation": DataStatus.RAW
        }

    def load_transcripts(self):
        """ Loads transcripts and translations. """
        if self._transc_transla_list is not None:
            return
        logging.info(f"Loading transcriptions from tarball: {self._input_tarball}")
        with self.open_tarball("zip") as fp:
            for n in fp.namelist():
                if n.endswith("alignments.meta"):
                    mapping = fp.read(n).decode("utf-8")
                    mapping_id = [line.strip().split()[4]
                                  for line in mapping.split("\n")[1:] if line.strip()]
                elif n.endswith(".en"):
                    transcript = fp.read(n).decode("utf-8")
                    transcript = transcript.strip().split("\n")
                elif n.endswith(".fr") and "gtrans" not in n:
                    translation = fp.read(n).decode("utf-8")
                    translation = translation.strip().split("\n")
                elif "gtrans" in n:
                    g_translation = fp.read(n).decode("utf-8")
                    g_translation = g_translation.strip().split("\n")

        trans = [(idx, tc.strip().lower(), ts.strip(), gts.strip()) for idx, tc, ts, gts
                 in zip(mapping_id, transcript, translation, g_translation)
                 if tc.strip() and ts.strip() and gts.strip()]
        # [uttid, transcript, translation, google-translation]
        self._transc_transla_list = trans
        self._transcripts = [x[1] for x in trans]
        self._translations = [x[2] for x in trans]
        if "train" in os.path.basename(self._input_tarball):
            self._transcripts.extend([x[1] for x in trans])
            self._translations.extend(x[3] for x in trans)
        logging.info("Total {} utterances.".format(len(self._transcripts)))

    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
        """

        def gen():
            if self._transc_transla_list is None:
                self.load_transcripts()
            uttid2wavname = dict()
            with self.open_tarball("zip") as fp:
                for name in fp.namelist():
                    if name.endswith(".wav"):
                        uttid2wavname[name.split("/")[-1].split(".")[0]] = name
                n = 0
                for uttid, transcript, transla, g_transla in self._transc_transla_list:
                    iter_list = [transla]
                    if "train" in os.path.basename(self._input_tarball):
                        iter_list.append(g_transla)
                    for translation in iter_list:
                        n += 1
                        if total_shards > 1:
                            if n % total_shards != shard_id:
                                continue
                        binary_data = fp.read(uttid2wavname[uttid])
                        b = io.BytesIO(binary_data)
                        audio = self.extract_audio_feature(fileobj=b, mode="wav")
                        b.close()
                        data_sample = {
                            "audio": audio,
                            "transcript": transcript,
                            "translation": translation,
                        }
                        if map_func is None:
                            yield data_sample
                        else:
                            yield map_func(data_sample)

        return gen



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from neurst.data.text.jieba_segment import Jieba


def test():
    tok = Jieba()
    assert tok.tokenize("他来到了网易杭研大厦", return_str=True) == "他 来到 了 网易 杭研 大厦"
    assert tok.detokenize("他 来到 了 网易 杭研 大厦", return_str=True) == "他来到了网易杭研大厦"


if __name__ == "__main__":
    test()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import app, logging

import neurst.utils.flags_core as flags_core
from neurst.data.datasets import Dataset, build_dataset
from neurst.data.datasets.audio.audio_dataset import AudioTripleTFRecordDataset
from neurst.data.datasets.parallel_text_dataset import ParallelTextDataset
from neurst.exps import build_exp
from neurst.exps.sequence_generator import SequenceGenerator
from neurst.layers.search import SequenceSearch
from neurst.metrics.metric import Metric
from neurst.tasks import build_task
from neurst.training import training_utils
from neurst.utils.configurable import ModelConfigs

FLAG_LIST = [
    flags_core.Flag("distribution_strategy", dtype=flags_core.Flag.TYPE.STRING, default="mirrored",
                    help="The distribution strategy."),
    flags_core.Flag("dtype", dtype=flags_core.Flag.TYPE.STRING, default="float16",
                    help="The computation type of the whole model."),
    flags_core.Flag("enable_check_numerics", dtype=flags_core.Flag.TYPE.BOOLEAN, default=None,
                    help="Whether to open the tf.debugging.enable_check_numerics. "
                         "Note that this may lower down the training speed."),
    flags_core.Flag("asr_model_dir", dtype=flags_core.Flag.TYPE.STRING,
                    help="The path to the ASR model checkpoint."),
    flags_core.Flag("mt_model_dir", dtype=flags_core.Flag.TYPE.STRING,
                    help="The path to the MT model checkpoint."),
    flags_core.Flag("asr_output_file", dtype=flags_core.Flag.TYPE.STRING,
                    help="The path to save ASR hypothesis."),
    flags_core.Flag("mt_output_file", dtype=flags_core.Flag.TYPE.STRING,
                    help="The path to save MT hypothesis."),
    flags_core.Flag("batch_size", dtype=flags_core.Flag.TYPE.INTEGER, default=32,
                    help="The batch size for inference."),
    flags_core.ModuleFlag(Dataset.REGISTRY_NAME, default=AudioTripleTFRecordDataset.__name__,
                          help="The audio dataset."),
    flags_core.ModuleFlag("asr_" + SequenceSearch.REGISTRY_NAME,
                          module_name=SequenceSearch.REGISTRY_NAME,
                          default="beam_search", help="The search method for ASR."),
    flags_core.ModuleFlag("mt_" + SequenceSearch.REGISTRY_NAME,
                          module_name=SequenceSearch.REGISTRY_NAME,
                          default="beam_search", help="The search method for MT."),
    flags_core.ModuleFlag("asr_" + Metric.REGISTRY_NAME,
                          module_name=Metric.REGISTRY_NAME,
                          default="wer", help="The metric to evaluate ASR output."),
    flags_core.ModuleFlag("mt_" + Metric.REGISTRY_NAME,
                          module_name=Metric.REGISTRY_NAME,
                          default="bleu", help="The metric to evaluate MT output."),
]


def _build_task_model(strategy, model_dir, batch_size):
    with training_utils.get_strategy_scope(strategy):
        model_configs = ModelConfigs.load(model_dir)
        task = build_task(model_configs, batch_size=batch_size)
        model = task.build_model(model_configs)
        return task, model


def _main(_):
    # define and parse program flags
    arg_parser = flags_core.define_flags(FLAG_LIST)
    args, remaining_argv = flags_core.parse_flags(FLAG_LIST, arg_parser)
    flags_core.verbose_flags(FLAG_LIST, args, remaining_argv)
    strategy = training_utils.handle_distribution_strategy(args["distribution_strategy"])
    training_utils.startup_env(dtype=args["dtype"], enable_xla=False,
                               enable_check_numerics=args["enable_check_numerics"])

    asr_task, asr_model = _build_task_model(strategy, args["asr_model_dir"],
                                            batch_size=args["batch_size"])
    mt_task, mt_model = _build_task_model(strategy, args["mt_model_dir"],
                                          batch_size=args["batch_size"])
    audio_dataset = build_dataset(args)
    # ========= ASR ==========
    asr_output_file = args["asr_output_file"]
    if asr_output_file is None:
        asr_output_file = "ram://asr_output_file"
    logging.info("Creating ASR generator.")
    with training_utils.get_strategy_scope(strategy):
        asr_generator = build_exp(
            {"class": SequenceGenerator,
             "params": {
                 "output_file": asr_output_file,
                 "search_method.class": args["asr_search_method.class"],
                 "search_method.params": args["asr_search_method.params"],
             }},
            strategy=strategy,
            model=asr_model,
            task=asr_task,
            model_dir=args["asr_model_dir"],
            custom_dataset=audio_dataset)
    asr_generator.run()
    if hasattr(audio_dataset, "transcripts") and audio_dataset.transcripts is not None:
        asr_metric = asr_task.get_eval_metric(args, "asr_metric")
        with tf.io.gfile.GFile(asr_output_file, "r") as fp:
            metric_result = asr_metric([line.strip() for line in fp],
                                       audio_dataset.transcripts)
        logging.info("Evaluation Result of ASR:")
        for k, v in metric_result.items():
            logging.info("   %s=%.2f", k, v)

    logging.info("Creating MT generator.")
    mt_reference_file = "ram://mt_reference_file"
    with tf.io.gfile.GFile(mt_reference_file, "w") as fw:
        for x in audio_dataset.targets:
            fw.write(x.strip() + "\n")

    with training_utils.get_strategy_scope(strategy):
        mt_generator = build_exp(
            {"class": SequenceGenerator,
             "params": {
                 "output_file": args["mt_output_file"],
                 "search_method.class": args["mt_search_method.class"],
                 "search_method.params": args["mt_search_method.params"],
                 "metric.class": args["mt_metric.class"],
                 "metric.params": args["mt_metric.params"]
             }},
            strategy=strategy,
            model=mt_model,
            task=mt_task,
            model_dir=args["mt_model_dir"],
            custom_dataset=build_dataset({
                "class": ParallelTextDataset,
                "params": {
                    "src_file": asr_output_file,
                    "trg_file": mt_reference_file
                }}))
    mt_generator.run()


def cli_main():
    logging.set_verbosity(logging.INFO)
    app.run(_main, argv=["pseudo.py"])


if __name__ == "__main__":
    cli_main()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy
import tensorflow as tf


def gelu(x, non_approximate=False):
    """Gaussian Error Linear Unit.
    This is a smoother version of the RELU.
    Original paper: https://arxiv.org/abs/1606.08415
    Args:
        x: float Tensor to perform activation.
        non_approximate: use tanh approximation
    Returns:
        `x` with the GELU activation applied.
    """
    if non_approximate:
        # TODO: check fp16
        # https://github.com/tensorflow/tensorflow/issues/25052
        if x.dtype.base_dtype.name == "float16":
            fp32_x = tf.cast(x, tf.float32)
        else:
            fp32_x = x
        cdf = 0.5 * (1.0 + tf.math.erf(fp32_x / numpy.sqrt(2.0)))

        if x.dtype.base_dtype.name == "float16":
            return x * tf.saturate_cast(cdf, tf.float16)

        return x * cdf
    cdf = 0.5 * (1.0 + tf.tanh(
        (numpy.sqrt(2 / numpy.pi) * (x + 0.044715 * tf.pow(x, 3)))))
    return x * cdf


def glu(x):
    """ Gated linear unit. """
    a, b = tf.split(x, axis=-1, num_or_size_splits=2)
    return a * tf.sigmoid(b)


def get_activation(activ):
    if callable(activ):
        return activ
    if activ is None:
        return None
    if activ == "tanh":
        return tf.nn.tanh
    elif activ == "relu":
        return tf.nn.relu
    elif activ == "gelu":
        return lambda x: gelu(x, non_approximate=False)
    elif activ == "gelu_nonapprox":
        return lambda x: gelu(x, non_approximate=True)
    else:
        raise ValueError("Unknown activation: {}".format(activ))



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import multiprocessing
import os
from urllib.request import urlretrieve

import numpy
import tensorflow as tf
from absl import logging


def deprecated(obj):
    """This is a decorator which can be used to mark functions or classes
    as deprecated. It will result in a warning being emmitted
    when the function/class is used."""

    def new_obj(*args, **kwargs):
        logging.info("WARNING: Calling deprecated function/class: %s." % obj.__name__)
        return obj(*args, **kwargs)

    return new_obj


def flatten_string_list(arg):
    """ Flattens a string list.

    Args:
        arg: A list of string or a string. The string may contains comma as a separator.

    Returns: A list of string.
    """
    if arg is None:
        return None
    return [c.strip() for cs in tf.nest.flatten(arg) for c in cs.split(",")]


class DummyContextManager(object):

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass


def to_numpy_or_python_type(tensors, bytes_as_str=False):
    """Converts a structure of `Tensor`s to `NumPy` arrays or Python scalar types.

    For each tensor, it calls `tensor.numpy()`. If the result is a scalar value,
    it converters it to a Python type, such as a float or int, by calling
    `result.item()`.

    Numpy scalars are converted, as Python types are often more convenient to deal
    with. This is especially useful for bfloat16 Numpy scalars, which don't
    support as many operations as other Numpy values.

    Args:
      tensors: A structure of tensors.
      bytes_as_str: A boolean, whether to convert bytes elements to string.

    Returns:
      `tensors`, but scalar tensors are converted to Python types and non-scalar
      tensors are converted to Numpy arrays.
    """

    def _to_single_numpy_or_python_type(t):
        if isinstance(t, tf.Tensor):
            x = t.numpy()
            if numpy.ndim(x) == 0:
                x = x.item()
                if bytes_as_str and isinstance(x, bytes):
                    return x.decode("utf-8")
                return x
            else:
                if len(x) == 0:
                    return x
                if bytes_as_str and isinstance(x.flatten()[0], bytes):
                    return tf.nest.map_structure(lambda _x: _x.decode("utf-8"), x.tolist())[0]
                return x
        return t  # Don't turn ragged or sparse tensors to NumPy.

    return tf.nest.map_structure(_to_single_numpy_or_python_type, tensors)


class PseudoPool(object):
    def __init__(self, processes=1):
        """ If processes is 1, then don't create pool.

        Args:
            processes:
        """
        self.pool = None
        if processes > 1:
            self.pool = multiprocessing.Pool(processes=processes)
        self.processes = processes

    @staticmethod
    def parse_arg_list(n_threads, sample_list, *args):
        total_length = len(sample_list)
        samples_per_thread = total_length // n_threads + 1
        if samples_per_thread < 1000:
            samples_per_thread = total_length
        start_idx = 0
        ret = []
        while start_idx < total_length:
            arg_list = []
            arg_list.append(sample_list[start_idx:(start_idx + samples_per_thread)])
            for arg in args:
                if isinstance(arg, list) and len(arg) == len(sample_list):
                    arg_list.append(arg[start_idx:(start_idx + samples_per_thread)])
                elif not isinstance(arg, list):
                    arg_list.append(arg)
                else:
                    raise ValueError
            start_idx += samples_per_thread
            ret.append(arg_list)
        return ret

    def map(self, func, args_list):
        return [func(args) for args in args_list]

    def __enter__(self):
        if self.processes > 1:
            return self.pool
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.processes > 1:
            self.pool.terminate()


def download_with_tqdm(url, filename):
    from tqdm import tqdm

    class TqdmUpTo(tqdm):
        last_block = 0

        def update_to(self, block_num=1, block_size=1, total_size=None):
            if total_size is not None:
                self.total = total_size
            self.update((block_num - self.last_block) * block_size)
            self.last_block = block_num

    try:
        os.remove(filename)
    except FileNotFoundError:
        pass

    with TqdmUpTo(unit="B", unit_scale=True, unit_divisor=1024,
                  miniters=1, desc=filename) as t:
        urlretrieve(url, filename, reporthook=t.update_to, data=None)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Defines thai tokenizer."""
from neurst.data.text import Tokenizer, register_tokenizer


@register_tokenizer
class ThaiTokenizer(Tokenizer):

    def __init__(self, language="th", glossaries=None):
        """ Initializes. """
        _ = language
        language = "th"
        try:
            from thai_segmenter import tokenize as thai_tokenize
            self._thai_tokenize = thai_tokenize
        except ImportError:
            raise ImportError('Please install Thai tokenizer with: pip install thai-segmenter')
        self._thai_tokenize("")
        super(ThaiTokenizer, self).__init__(language=language, glossaries=glossaries)

    def tokenize(self, text, return_str=False):
        """ Tokenize a text. """
        res = self._thai_tokenize(self._convert_to_str(text))
        if return_str:
            res = " ".join(res)
        return res

    def detokenize(self, words, return_str=True):
        """ Recovers the result of `tokenize(words)`.

        Args:
            words: A list of strings, i.e. tokenized text.
            return_str: returns a string if True, a list of tokens otherwise.

        Returns: The recovered sentence string.
        """
        if isinstance(words, str):
            words = words.strip().split()
        if return_str:
            words = "".join(words)
        return words



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from neurst.data.text import register_tokenizer
from neurst.data.text.tokenizer import Tokenizer


@register_tokenizer("char")
class Character(Tokenizer):
    CHAR_COMPILER = re.compile(r"([\u2E80-\u9FFF\uA000-\uA4FF\uAC00-\uD7FF\uF900-\uFAFF])")

    def __init__(self, language, glossaries=None):
        super(Character, self).__init__(
            language=language, glossaries=glossaries)

    @staticmethod
    def is_cjk(language):
        return language in ["zh", "ja", "ko"]

    @staticmethod
    def to_character(text, language=None):
        text = Character._convert_to_str(text)
        if Character.is_cjk(language):
            return Character.cjk_to_character(text)
        return " ".join(text)

    @staticmethod
    def cjk_to_character(text):
        """ CJK sentence to character-level.

        Args:
            text: A list of string tokens or a string.

        Returns: A string.
        """
        text = Character._convert_to_str(text)
        res = Character.CHAR_COMPILER.sub(r" \1 ", text)
        # tokenize period and comma unless preceded by a digit
        res = re.sub(r'([^0-9])([\.,])', r'\1 \2 ', res)

        # tokenize period and comma unless followed by a digit
        res = re.sub(r'([\.,])([^0-9])', r' \1 \2', res)

        # tokenize dash when preceded by a digit
        res = re.sub(r'([0-9])(-)', r'\1 \2 ', res)

        # one space only between words
        res = re.sub(r'\s+', r' ', res)

        # no leading space
        res = re.sub(r'^\s+', r'', res)

        # no trailing space
        res = re.sub(r'\s+$', r'', res)
        return res

    def tokenize(self, text, return_str=False):
        return self._output_wrapper(
            self.to_character(text, language=self.language),
            return_str=return_str)

    def detokenize(self, text, return_str=True):
        if not self.is_cjk(self.language):
            raise NotImplementedError(f"detokenize fn in Character "
                                      f"is not implemented for language={self.language}")
        return self._output_wrapper(
            self.cjk_deseg(self._convert_to_str(text)),
            return_str=return_str)



import importlib
import os

from neurst.data.audio.feature_extractor import FeatureExtractor
from neurst.utils.registry import setup_registry

build_feature_extractor, register_feature_extractor = setup_registry(
    "feature_extractor", base_class=FeatureExtractor, verbose_creation=True)

models_dir = os.path.dirname(__file__)
for file in os.listdir(models_dir):
    path = os.path.join(models_dir, file)
    if not file.startswith('_') and not file.startswith('.') and file.endswith('.py'):
        model_name = file[:file.find('.py')] if file.endswith('.py') else file
        module = importlib.import_module('neurst.data.audio.' + model_name)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Implements transformer decoder in TF2 as described in https://arxiv.org/abs/1706.03762. """
import tensorflow as tf

from neurst.layers import build_transformer_component, layer_utils
from neurst.layers.attentions.multi_head_attention import MultiHeadAttention, MultiHeadSelfAttention
from neurst.layers.common_layers import TransformerFFN
from neurst.layers.decoders import Decoder, register_decoder
from neurst.utils import compat


@register_decoder
class TransformerDecoder(Decoder):
    """ Defines transformer encoder as described
    in https://arxiv.org/abs/1706.03762. """

    def __init__(self,
                 num_layers,
                 hidden_size,
                 num_attention_heads,
                 filter_size,
                 ffn_activation="relu",
                 attention_dropout_rate=0.,
                 attention_type="dot_product",
                 ffn_dropout_rate=0.,
                 layer_postprocess_dropout_rate=0.,
                 layer_postprocess_epsilon=1e-6,
                 with_encoder_decoder_attention=True,
                 post_normalize=False,
                 name=None):
        """ Initializes the parameters of the transformer decoder.

        Args:
            num_layers: The number of stacked layers.
            hidden_size: The number of hidden units.
            num_attention_heads: The number of self attention heads, for both self-attention
                and encoder-decoder attention.
            filter_size: The filter size of ffn layer.
            ffn_activation: The activation function of ffn layer.
            ffn_dropout_rate: The dropout rate for ffn layer.
            attention_dropout_rate: The dropout rate for ffn layer, for both self-attention
                and encoder-decoder attention.
            attention_type: The self attention type, for both self-attention and
                encoder-decoder attention.
            layer_postprocess_dropout_rate: The dropout rate for each
                layer post process.
            layer_postprocess_epsilon: The epsilon for layer norm.
            post_normalize: Whether to apply layernorm after each block.
            name: The name of this decoder.
        """
        super(TransformerDecoder, self).__init__(
            num_layers=num_layers, hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            filter_size=filter_size, ffn_activation=ffn_activation,
            ffn_dropout_rate=ffn_dropout_rate,
            attention_type=attention_type,
            attention_dropout_rate=attention_dropout_rate,
            layer_postprocess_dropout_rate=layer_postprocess_dropout_rate,
            layer_postprocess_epsilon=layer_postprocess_epsilon,
            with_encoder_decoder_attention=with_encoder_decoder_attention,
            post_normalize=post_normalize,
            name=name or self.__class__.__name__)
        self._stacking_layers = []
        self._with_encoder_decoder_attention = with_encoder_decoder_attention

    def build(self, input_shape):
        """ Builds the transformer decoder layer. """
        params = self.get_config()
        for _ in range(params["num_layers"]):
            self._stacking_layers.append([
                build_transformer_component({
                    "base_layer.class": MultiHeadSelfAttention.__name__,
                    "base_layer.params": dict(
                        num_heads=params["num_attention_heads"],
                        num_units=params["hidden_size"],
                        attention_dropout_rate=params["attention_dropout_rate"],
                        attention_type=params["attention_type"],
                        name="self_attention"
                    )},
                    dropout_rate=params["layer_postprocess_dropout_rate"],
                    epsilon=params["layer_postprocess_epsilon"],
                    pre_norm=(not params["post_normalize"])),
                (build_transformer_component({
                    "base_layer.class": MultiHeadAttention.__name__,
                    "base_layer.params": dict(
                        num_heads=params["num_attention_heads"],
                        num_units=params["hidden_size"],
                        attention_dropout_rate=params["attention_dropout_rate"],
                        attention_type=params["attention_type"],
                        name="encdec_attention")},
                    dropout_rate=params["layer_postprocess_dropout_rate"],
                    epsilon=params["layer_postprocess_epsilon"],
                    pre_norm=(not params["post_normalize"]))
                 if self._with_encoder_decoder_attention else None),
                build_transformer_component({
                    "base_layer.class": TransformerFFN.__name__,
                    "base_layer.params": dict(
                        filter_size=params["filter_size"],
                        output_size=params["hidden_size"],
                        dropout_rate=params["ffn_dropout_rate"],
                        activation=params["ffn_activation"],
                        name="ffn")},
                    dropout_rate=params["layer_postprocess_dropout_rate"],
                    epsilon=params["layer_postprocess_epsilon"],
                    pre_norm=(not params["post_normalize"]))])
        if not params["post_normalize"]:
            self._output_norm_layer = tf.keras.layers.LayerNormalization(
                epsilon=params["layer_postprocess_epsilon"],
                dtype="float32", name="output_ln")
        super(TransformerDecoder, self).build(input_shape)

    def create_decoding_internal_cache(self,
                                       encoder_outputs,
                                       encoder_inputs_padding,
                                       is_inference=False,
                                       decode_padded_length=None):
        """ Creates internal cache for decoding.

        Args:
            encoder_outputs: The output tensor from encoder
                with shape [batch_size, max_input_length, hidden_size].
            encoder_inputs_padding: A float tensor with shape [batch_size, max_length],
                indicating the padding positions of `encoder_output`, where 1.0 for
                padding and 0.0 for non-padding.
            is_inference: A boolean scalar, whether in inference mode or not.
            decode_padded_length: The maximum decoding length when inference, for creating
                static-shape cache.

        Returns:
            `cache`, a dictionary containing static(e.g. encoder hidden states
            for attention) and dynamic(e.g. transformer decoding cache) tensors used
            during decoding and will be passed to `call()`. Note that, the dynamic
            tensors must store in cache["decoding_states"] for beam search use.
        """
        # [batch_size, max_length], FLOAT_MIN for padding, 0.0 for non-padding
        if is_inference:
            params = self.get_config()
            decoding_states = {}
            batch_size = tf.shape(encoder_outputs)[0]
            num_heads = params["num_attention_heads"]
            num_units_per_head = params["hidden_size"] // num_heads
            # initialize decoder self attention keys/values
            for lid in range(params["num_layers"]):
                # Ensure shape invariance for tf.while_loop.
                decoding_states["layer_{}".format(lid)] = {
                    "self_attention": {
                        "keys": tf.zeros([batch_size, decode_padded_length or 0, num_heads, num_units_per_head],
                                         dtype=compat.CUSTOM_GLOBAL_FLOATX),
                        "values": tf.zeros([batch_size, decode_padded_length or 0, num_heads, num_units_per_head],
                                           dtype=compat.CUSTOM_GLOBAL_FLOATX)},
                }
        else:
            decoding_states = None
        cache = dict(decoding_states=decoding_states)
        if self._with_encoder_decoder_attention:
            cache["memory"] = encoder_outputs
            cache["memory_bias"] = layer_utils.input_padding_to_bias(
                encoder_inputs_padding)
        return cache

    def call(self, decoder_inputs, cache, is_training=True, decode_loop_step=None):
        """ Encodes the inputs.

        Args:
            decoder_inputs: The embedded decoder input, a float tensor with shape
                [batch_size, max_target_length, embedding_dim] or
                [batch_size, embedding_dim] for one decoding step.
            cache: A dictionary, generated from self.create_decoding_internal_cache.
            is_training: A bool, whether in training mode or not.
            decode_loop_step: An integer, step number of the decoding loop. Used only
                for autoregressive inference with static-shape cache.

        Returns:
            The decoder output with shape [batch_size, max_length, hidden_size]
            when `decoder_inputs` is a 3-d tensor or with shape
            [batch_size, hidden_size] when `decoder_inputs` is a 2-d tensor.
        """
        ori_ndims = decoder_inputs.get_shape().ndims
        if ori_ndims == 2:
            decoder_inputs = tf.expand_dims(decoder_inputs, axis=1)

        # decoder self attention has shape [1, 1, max_target_len, max_target_len]
        decoder_self_attention_bias = layer_utils.lower_triangle_attention_bias(
            tf.shape(decoder_inputs)[1])
        x = decoder_inputs
        if is_training:
            x = tf.nn.dropout(
                decoder_inputs, rate=self.get_config()["layer_postprocess_dropout_rate"])
        for idx, layer in enumerate(self._stacking_layers):
            selfatt_layer = layer[0]
            encdecatt_layer = layer[1]
            ffn_layer = layer[2]
            layer_name = "layer_{}".format(idx)
            layer_cache = None if cache["decoding_states"] is None else cache["decoding_states"][layer_name]
            selfatt_cache = None if layer_cache is None else layer_cache["self_attention"]
            with tf.name_scope(layer_name):
                # self attention layer
                x = selfatt_layer(
                    x,  # x as query
                    bias=decoder_self_attention_bias,
                    cache=selfatt_cache,
                    is_training=is_training,
                    decode_loop_step=decode_loop_step)
                # enc-dec attention layer
                if encdecatt_layer is not None:
                    x = encdecatt_layer(
                        x,  # x as query
                        memory=cache["memory"],  # None indicates self-attention
                        memory_bias=cache["memory_bias"],
                        is_training=is_training)
                # ffn
                x = ffn_layer(x, is_training=is_training)
        if not self.get_config()["post_normalize"]:
            outputs = self._output_norm_layer(x)
        if ori_ndims == 2:
            outputs = tf.squeeze(outputs, axis=1)
        return outputs



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import copy
import json
from collections import namedtuple

from absl import logging

from neurst.utils.configurable import deep_merge_dict, load_from_config_path, yaml_load_checking
from neurst.utils.misc import flatten_string_list
from neurst.utils.registry import REGISTRIES

_DEFINED_FLAGS = dict()


class Flag(object):
    TYPE = namedtuple(
        "FLAG_ARG_TYPES", "INTEGER BOOLEAN FLOAT STRING")(int, bool, float, str)

    UNIQ_SET = []

    def __init__(self,
                 name,
                 dtype,
                 required=False,
                 choices=None,
                 help="",
                 default=None,
                 multiple=False,
                 alias=None):
        """ The flags

        Args:
            name: The flag name.
            dtype: The type.
            choices: A list of acceptable values.
            default: The default value.
            help: The help text.
            multiple: Whether the flag accepts multiple arguments.
            alias: The alias name for this flag.
        """
        if name in ["class", "params"]:
            raise ValueError("Invalid flag name: {}".format(name))
        if "-" in name:
            raise ValueError("Flag name with '-' is not supported.")
        self._name = name.strip()
        self._dtype = dtype
        self._default = default
        self._help = help
        self._choices = choices
        self._multiple = multiple
        self._required = required
        self._alias = alias

    @property
    def name(self):
        return self._name

    @property
    def dtype(self):
        return self._dtype

    @property
    def default(self):
        return self._default

    @property
    def multiple(self):
        return self._multiple

    @property
    def help(self):
        return self._help

    @property
    def alias(self):
        return self._alias

    @property
    def choices(self):
        return self._choices

    def define(self, arg_parser: argparse.ArgumentParser, default_is_none=True):
        """ Adds argument to the parser.

        Args:
            arg_parser: An ArgumentParser object.
            default_is_none: Whether the default value is None

        Returns: The parser.
        """
        try:
            idx = self.UNIQ_SET.index(self)
        except ValueError:
            self.UNIQ_SET.append(self)
        else:
            raise ValueError("Defines duplicate flag: {}, while {} is already exists.".format(
                str(self), str(self.UNIQ_SET[idx])))

        flag_names = ["--" + self.name]
        if self.alias is not None:
            flag_names.append("--" + self.alias)
        kwargs = {"type": self.dtype, "dest": self.name, "help": self.help}
        if self.dtype is bool:
            kwargs["action"] = "store_true"
            kwargs["default"] = None
            kwargs.pop("type")
        if self.multiple:
            kwargs["nargs"] = "+"
        if self.choices:
            kwargs["choices"] = self.choices
        if self.default and not default_is_none:
            kwargs["default"] = self.default
        if self._required:
            kwargs["required"] = True
        arg_parser.add_argument(*flag_names, **kwargs)
        if self.name in _DEFINED_FLAGS:
            raise ValueError(f"Defined duplicate arg key: {self.name}")
        _DEFINED_FLAGS[self.name] = self
        return arg_parser


class ModuleFlag(object):
    def __init__(self, name, module_name=None, default=None, help=""):
        """ Initializes the module flag. """
        self._name = name
        self._module_name = module_name or name
        self._help = help
        self._default = default

    @property
    def help(self):
        return self._help

    @property
    def default(self):
        return self._default

    @property
    def name(self):
        return self._name

    @property
    def module_name(self):
        return self._module_name

    @property
    def cls_key(self):
        return self.name + ".class"

    @property
    def params_key(self):
        return self.name + ".params"

    def define(self, arg_parser: argparse.ArgumentParser):
        """ Adds argument to the parser.

        Args:
            arg_parser: An ArgumentParser object.

        Returns: The parser.
        """
        _DEFINED_FLAGS[self.name] = self
        Flag(name=self.cls_key, dtype=Flag.TYPE.STRING, alias=self.name,
             choices=list(REGISTRIES[self.module_name].keys()), default=self.default,
             help=f"The class name of {self.module_name} for '{self.help}'").define(
            arg_parser, default_is_none=False)
        Flag(name=self.params_key, dtype=Flag.TYPE.STRING, default="{}",
             help=f"The json/yaml-like parameter string for {self.module_name}").define(
            arg_parser, default_is_none=False)
        return arg_parser


COMMON_DATA_ARGS = [
    Flag("shuffle_buffer", dtype=Flag.TYPE.INTEGER, default=0,
         help="The buffer size for dataset shuffle."),
    Flag("batch_size", dtype=Flag.TYPE.INTEGER, default=None,
         help="The number of samples per update."),
    Flag("batch_size_per_gpu", dtype=Flag.TYPE.INTEGER, default=None,
         help="The per-GPU batch size, that takes precedence of `batch_size`."),
    Flag("cache_dataset", dtype=Flag.TYPE.BOOLEAN,
         help="Whether to cache the training data in memory.")]

DEFAULT_CONFIG_FLAG = Flag(name="config_paths", dtype=Flag.TYPE.STRING, multiple=True,
                           help="Path to a json/yaml configuration files defining FLAG values. "
                                "Multiple files can be separated by commas. Files are merged recursively. "
                                "Setting a key in these files is equivalent to "
                                "setting the FLAG value with the same name.")


def define_flags(flag_list: list, arg_parser=None, with_config_file=True) -> argparse.ArgumentParser:
    """ Defines the root module name.

    Args:
        flag_list: A list of flags.
        arg_parser: The argument parser.
        with_config_file: Whether to define `config_paths` as default.

    Returns: The argument parser.
    """
    if arg_parser is None:
        arg_parser = argparse.ArgumentParser()
    if with_config_file:
        DEFAULT_CONFIG_FLAG.define(arg_parser)
    for f in flag_list:
        f.define(arg_parser)
    return arg_parser


def get_argparser(module_name, cls_name) -> argparse.ArgumentParser:
    """ Returns the argument parser for the class.

    Args:
        module_name: The registered module name.
        cls_name: The class name (or alias).

    Returns: An argument parser that parses the class args.
    """
    arg_parser = argparse.ArgumentParser()
    if hasattr(REGISTRIES[module_name][cls_name], "class_or_method_args"):
        for f in REGISTRIES[module_name][cls_name].class_or_method_args():
            f.define(arg_parser)
    return arg_parser


def _flatten_args(flag_list, from_args):
    args = copy.deepcopy(from_args)
    flattened_args = {}
    for f in flag_list:
        if isinstance(f, Flag) and f.name in args:
            flattened_args[f.name] = args.pop(f.name)
    for f in flag_list:
        if isinstance(f, ModuleFlag):
            if f.cls_key in args:
                flattened_args[f.cls_key] = args.pop(f.cls_key)
                args.pop(f.name, None)
            elif f.name in args:
                flattened_args[f.cls_key] = args.pop(f.name)
            if f.cls_key in flattened_args and flattened_args[f.cls_key] and args.get(f.params_key, None):
                if hasattr(REGISTRIES[f.module_name][flattened_args[f.cls_key]], "class_or_method_args"):
                    for ff in REGISTRIES[f.module_name][flattened_args[f.cls_key]].class_or_method_args():
                        if isinstance(ff, Flag):
                            if ff.name in args[f.params_key] and ff.name not in flattened_args:
                                flattened_args[ff.name] = args[f.params_key].pop(ff.name)
                        else:
                            if ff.cls_key in args:
                                flattened_args[ff.cls_key] = args.pop(ff.cls_key)
                                args.pop(ff.name, None)
                            elif ff.name in args:
                                flattened_args[ff.cls_key] = args.pop(ff.name)
                            elif ff.cls_key in args[f.params_key]:
                                flattened_args[ff.cls_key] = args[f.params_key].pop(ff.cls_key)
                            elif ff.name in args[f.params_key]:
                                flattened_args[ff.cls_key] = args[f.params_key].pop(ff.name)
                            if ff.params_key in args[f.params_key]:
                                flattened_args[ff.params_key] = deep_merge_dict(
                                    args[f.params_key][ff.params_key], flattened_args.get(ff.params_key, {}))
                else:
                    flattened_args[f.params_key] = args.pop(f.params_key)
                args.pop(f.params_key, None)
    return deep_merge_dict(flattened_args, args)


def _args_preload_from_config_files(args):
    cfg_file_args = yaml_load_checking(load_from_config_path(
        flatten_string_list(getattr(args, DEFAULT_CONFIG_FLAG.name, None))))
    return cfg_file_args


def parse_flags(flag_list, arg_parser: argparse.ArgumentParser,
                args_preload_func=_args_preload_from_config_files):
    """ Parses flags from argument parser.

    Args:
        flag_list: A list of flags.
        arg_parser: The program argument parser.
        args_preload_func: A callable function for pre-loading arguments, maybe from
            config file, hyper parameter set.
    """
    program_parsed_args, remaining_argv = arg_parser.parse_known_args()
    cfg_file_args = {}
    if args_preload_func is not None:
        cfg_file_args = args_preload_func(program_parsed_args)
    program_parsed_args = yaml_load_checking(program_parsed_args.__dict__)
    top_program_parsed_args = {}
    for f in flag_list:
        flag_key = f.name
        if isinstance(f, ModuleFlag):
            flag_key = f.cls_key
            top_program_parsed_args[f.params_key] = {}
            if program_parsed_args.get(f.params_key, None) is not None:
                top_program_parsed_args[f.params_key] = program_parsed_args[f.params_key]
            if f.params_key in cfg_file_args:
                top_program_parsed_args[f.params_key] = deep_merge_dict(
                    cfg_file_args[f.params_key], top_program_parsed_args[f.params_key])
        if program_parsed_args.get(flag_key, None) is not None:
            top_program_parsed_args[flag_key] = program_parsed_args[flag_key]
        elif flag_key in cfg_file_args:
            top_program_parsed_args[flag_key] = cfg_file_args[flag_key]
        else:
            top_program_parsed_args[flag_key] = f.default

    return top_program_parsed_args, remaining_argv


def intelligent_parse_flags(flag_list, arg_parser: argparse.ArgumentParser,
                            args_preload_func=_args_preload_from_config_files):
    """ Parses flags from argument parser.

    Args:
        flag_list: A list of flags.
        arg_parser: The program argument parser.
        args_preload_func: A callable function for pre-loading arguments, maybe from
            config file, hyper parameter set.
    """
    program_parsed_args, remaining_argv = arg_parser.parse_known_args()
    cfg_file_args = {}
    if args_preload_func is not None:
        cfg_file_args = args_preload_func(program_parsed_args)
    top_program_parsed_args = _flatten_args(flag_list,
                                            yaml_load_checking(program_parsed_args.__dict__))
    for f in flag_list:
        if isinstance(f, ModuleFlag):
            if f.cls_key in top_program_parsed_args and top_program_parsed_args[f.cls_key]:
                cfg_file_args[f.cls_key] = top_program_parsed_args[f.cls_key]
    cfg_file_args = _flatten_args(flag_list, cfg_file_args)
    for f in flag_list:
        if isinstance(f, Flag):
            if top_program_parsed_args[f.name] is None:
                top_program_parsed_args[f.name] = cfg_file_args.get(f.name, None)
            cfg_file_args.pop(f.name, None)
        else:
            submodule_cls = (top_program_parsed_args.get(f.cls_key, None)
                             or cfg_file_args.get(f.cls_key, None))
            cfg_file_args.pop(f.cls_key, None)
            if submodule_cls is None:
                continue
            top_program_parsed_args[f.cls_key] = submodule_cls
            if top_program_parsed_args.get(f.params_key, None) is None:
                top_program_parsed_args[f.params_key] = {}
            module_arg_parser = get_argparser(f.module_name, submodule_cls)
            module_parsed_args, remaining_argv = module_arg_parser.parse_known_args(remaining_argv)
            module_parsed_args = yaml_load_checking(module_parsed_args.__dict__)
            if hasattr(REGISTRIES[f.module_name][submodule_cls], "class_or_method_args"):
                cfg_file_args = _flatten_args(
                    REGISTRIES[f.module_name][submodule_cls].class_or_method_args(), cfg_file_args)
                for inner_f in REGISTRIES[f.module_name][submodule_cls].class_or_method_args():
                    flag_key = inner_f.name
                    if isinstance(inner_f, ModuleFlag):
                        flag_key = inner_f.cls_key
                    if module_parsed_args[flag_key] is not None:
                        top_program_parsed_args[f.params_key][flag_key] = module_parsed_args[flag_key]
                        top_program_parsed_args.pop(flag_key, None)
                        cfg_file_args.pop(flag_key, None)
                    elif flag_key in top_program_parsed_args:
                        top_program_parsed_args[f.params_key][flag_key] = top_program_parsed_args.pop(flag_key)
                        cfg_file_args.pop(flag_key, None)
                    elif flag_key in cfg_file_args:
                        top_program_parsed_args[f.params_key][flag_key] = cfg_file_args.pop(flag_key)

                    if isinstance(inner_f, ModuleFlag):
                        top_program_parsed_args[f.params_key][inner_f.params_key] = deep_merge_dict(
                            cfg_file_args.pop(inner_f.params_key, {}) or {},
                            deep_merge_dict(top_program_parsed_args.pop(inner_f.params_key, {}) or {},
                                            module_parsed_args.pop(inner_f.params_key, {}) or {}))
    top_program_parsed_args = deep_merge_dict(cfg_file_args, top_program_parsed_args)
    for f in flag_list:
        if isinstance(f, Flag):
            if f.name not in top_program_parsed_args or top_program_parsed_args[f.name] is None:
                top_program_parsed_args[f.name] = f.default
    return top_program_parsed_args, remaining_argv


def extend_define_and_parse(flag_name, args, remaining_argv):
    f = _DEFINED_FLAGS.get(flag_name, None)
    if f is None or not isinstance(f, ModuleFlag):
        return args
    if not hasattr(REGISTRIES[f.module_name][args[f.cls_key]], "class_or_method_args"):
        return args
    arg_parser = argparse.ArgumentParser()
    for ff in REGISTRIES[f.module_name][args[f.cls_key]].class_or_method_args():
        if isinstance(ff, ModuleFlag):
            if args[f.params_key].get(ff.cls_key, None):
                if hasattr(REGISTRIES[ff.module_name][args[f.params_key][ff.cls_key]], "class_or_method_args"):
                    for fff in REGISTRIES[ff.module_name][args[f.params_key][ff.cls_key]].class_or_method_args():
                        fff.define(arg_parser)
    parsed_args, remaining_argv = arg_parser.parse_known_args(remaining_argv)
    parsed_args = yaml_load_checking(parsed_args.__dict__)
    for ff in REGISTRIES[f.module_name][args[f.cls_key]].class_or_method_args():
        if isinstance(ff, ModuleFlag):
            if args[f.params_key].get(ff.cls_key, None):
                if hasattr(REGISTRIES[ff.module_name][args[f.params_key][ff.cls_key]], "class_or_method_args"):
                    if args[f.params_key].get(ff.params_key, None) is None:
                        args[f.params_key][ff.params_key] = {}
                    for fff in REGISTRIES[ff.module_name][args[f.params_key][ff.cls_key]].class_or_method_args():
                        flag_key = fff.name
                        if isinstance(fff, ModuleFlag):
                            flag_key = fff.cls_key
                        if parsed_args[flag_key] is not None:
                            args[f.params_key][ff.params_key][flag_key] = parsed_args[flag_key]
                            args.pop(flag_key, None)
                            args.pop(fff.name, None)
                        elif flag_key in args:
                            args[f.params_key][ff.params_key][flag_key] = args.pop(flag_key)
                            args.pop(fff.name, None)
                        elif fff.name in args:
                            args[f.params_key][ff.params_key][flag_key] = args.pop(fff.name)
                        elif fff.name in args[f.params_key][ff.params_key]:
                            if flag_name not in args[f.params_key][ff.params_key]:
                                args[f.params_key][ff.params_key][flag_key] = args[f.params_key][ff.params_key].pop(
                                    fff.name)
                        if isinstance(fff, ModuleFlag):
                            args[f.params_key][ff.params_key][fff.params_key] = deep_merge_dict(
                                args[f.params_key][ff.params_key].get(fff.params_key, {}) or {},
                                deep_merge_dict(args.get(fff.params_key, {}) or {},
                                                parsed_args.get(fff.params_key, {}) or {}))

    return args, remaining_argv


def verbose_flags(flag_list, args, remaining_argv):
    logging.info("==========================================================================")
    logging.info("Parsed all matched flags: ")
    verbose_args = copy.deepcopy(args)
    for f in flag_list:
        if isinstance(f, Flag):
            if f.name in verbose_args:
                logging.info(f" {f.name}: {verbose_args.pop(f.name)}     # (default: {f.default}) {f.help}")
        else:
            if f.cls_key in verbose_args:
                logging.info(f" {f.cls_key}: {verbose_args[f.cls_key]}")
            if f.params_key in verbose_args:
                if (verbose_args.get(f.cls_key, None) and hasattr(
                    REGISTRIES[f.module_name][verbose_args[f.cls_key]], "class_or_method_args")):
                    logging.info(f" {f.params_key}:")
                    for ff in REGISTRIES[f.module_name][verbose_args[f.cls_key]].class_or_method_args():
                        if isinstance(ff, Flag):
                            if ff.name in verbose_args[f.params_key]:
                                logging.info(f"   {ff.name}: {verbose_args[f.params_key][ff.name]}"
                                             f"     # (default: {ff.default}) {ff.help}")
                        else:
                            if ff.cls_key in verbose_args[f.params_key]:
                                logging.info(f"   {ff.cls_key}: {verbose_args[f.params_key][ff.cls_key]}")
                            if ff.params_key in verbose_args[f.params_key]:
                                logging.info(f"   {ff.params_key}: "
                                             f"{json.dumps(verbose_args[f.params_key][ff.params_key])}"
                                             f"     # {ff.help}")
                else:
                    logging.info(f" {f.params_key}: {json.dumps(verbose_args[f.params_key])}")
            verbose_args.pop(f.cls_key, None)
            verbose_args.pop(f.params_key, None)
    if len(verbose_args) > 0:
        logging.info("")
        logging.info("Other flags:")
        for k, v in verbose_args.items():
            logging.info(f" {k}: {str(v)}")
    if len(remaining_argv) > 0:
        logging.info("")
        logging.info("Remaining unparsed flags: ")
        text = None
        for arg in remaining_argv:
            if arg.startswith("-"):
                if text and len(text) > 0:
                    logging.info("  {}".format(" ".join(text)))
                text = []
            text.append(arg)
        if text and len(text) > 0:
            logging.info("  {}".format(" ".join(text)))
    logging.info("==========================================================================")
